SystemOrganization addCategory: #'Magma server-API'!
SystemOrganization addCategory: #'Magma server-Info'!
SystemOrganization addCategory: #'Magma server-Utilities'!
SystemOrganization addCategory: #'Magma server-private'!

!MagmaCollectionChanges methodsFor: '*magma server' stamp: 'cmm 4/29/2005 12:10'!
createLcMemberIndexUsing: aMaObjectRepository

	(MaHashIndex
		create: 
			(aMaObjectRepository 
				collectionFilenameFor: collectionOid
				extension: self suggestedFileExtension)
		keySize: MaObjectBuffer oidSize
		valueSize: 8
		recordSize: 10) close! !

!MagmaCollectionChanges methodsFor: '*magma server' stamp: 'cmm 3/17/2005 17:41'!
createNewLargeCollectionUsing: aMaObjectRepository
	"Create the necessary support-files for this type of index."

	self createLcMemberIndexUsing: aMaObjectRepository! !

!MagmaCollectionChanges methodsFor: '*magma server' stamp: 'cmm 3/17/2005 17:41'!
createSupplementaryIndexesUsing: aMaObjectRepository

	| mc |
	mc _ 
		aMaObjectRepository session 
			objectWithOid: collectionOid 
			ifAbsent:
				[ MagmaSoftwareError signal: 'expected to know about MagmaCollection ', collectionOid, ' here.' ].
	mc indexesDo:
		[ : eachIndexDefinition |
		(aMaObjectRepository ensureManagerCreatedFor: mc) ensureIndex: eachIndexDefinition ]! !

!MagmaCollectionChanges methodsFor: '*magma server' stamp: 'cmm 3/16/2005 15:13'!
requiresServerRefresh

	^ super requiresServerRefresh
		or: [ self hasAddedIndexes 
		or: [ self hasRemovedIndexes ] ]! !

!MagmaCollectionChanges methodsFor: '*magma server' stamp: 'cmm 3/20/2005 19:20'!
suggestedFileExtension

	^ MaHashIndex suggestedFileExtension! !

!MaVariableWordBuffer methodsFor: '*magma server' stamp: 'cmm 1/25/2005 14:07'!
appendGraphUsing: aMaReadStrategy into: aMaSerializedGraphBuffer currentDepth: currentDepth minDepth: minDepth with: alreadyAppendedSet filer: aMaObjectFiler

	"Nothing to do for a variable-word object."! !

!MaVariableWordBuffer methodsFor: '*magma server' stamp: 'cmm 5/1/2005 23:05'!
graphDo: oneArgBlock using: aMaReadStrategy for: aMagmaFileTraverser inDefinition: aBoolean

	"nothing to do for a word-object buffer."! !

!MagmaCollection methodsFor: '*magma server' stamp: 'cmm 3/17/2005 21:31'!
newManagerUsing: aMaObjectRepository

	^ (MagmaCollectionManager
		collection: self
		repository: aMaObjectRepository)
			oidOfIndexesCollection: (aMaObjectRepository session oidFor: self indexes) ;
			yourself! !

!FileStream methodsFor: '*magma server' stamp: 'cmm 7/15/2002 22:58'!
maRead: someNumber bytesFromPosition: bufferPosition of: aByteArray atFilePosition: filePosition

	^self
		position: filePosition;
		next: someNumber
		into: aByteArray
		startingAt: bufferPosition
! !

!FileStream methodsFor: '*magma server' stamp: 'cmm 7/15/2002 22:58'!
maWrite: numberOfBytes bytesFromPosition: bufferPosition of: byteArray atFilePosition: filePosition

	self position: filePosition.
	self
		next: numberOfBytes
		putAll: byteArray
		startingAt: bufferPosition.
	self maMarked: 'performance'.  "is flushing too slow?"
	"self flush"
! !

MaObject subclass: #MaLargeCollectionManager
	instanceVariableNames: 'collection memberIndex repository'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-private'!

!MaLargeCollectionManager commentStamp: 'cmm 3/16/2005 12:48' prior: 0!
I perform server-side operations for MagmaLargeCollections.  Adding, removing and accessing.!

!MaLargeCollectionManager class methodsFor: 'creation' stamp: 'cmm 3/17/2005 21:32'!
collection: aMagmaCollection repository: aMaObjectRepository

	^self new
		collection: aMagmaCollection;
		repository: aMaObjectRepository! !

!MaLargeCollectionManager methodsFor: 'initializing' stamp: 'cmm 3/16/2005 14:24'!
close

	memberIndex ifNotNil: [ memberIndex close.  memberIndex _ nil ]! !

!MaLargeCollectionManager methodsFor: 'private' stamp: 'cmm 3/17/2005 22:23'!
collection: aMagmaLargeCollection

	collection := aMagmaLargeCollection
! !

!MaLargeCollectionManager methodsFor: 'private' stamp: 'cmm 12/2/2004 13:22'!
collectionOid

	^ repository session oidFor: collection! !

!MaLargeCollectionManager methodsFor: 'initializing' stamp: 'cmm 3/19/2005 11:14'!
ensureMemberIndexOpen

	memberIndex ifNil:
		[ memberIndex _ self openIndex: 
			(repository
				collectionFilenameFor: self collectionOid
				extension: collection changes suggestedFileExtension) asString ]! !

!MaLargeCollectionManager methodsFor: 'actions' stamp: 'cmm 3/17/2005 17:22'!
ensureOpen

	^ self 
		ensureMemberIndexOpen ;
		yourself! !

!MaLargeCollectionManager methodsFor: 'accessing' stamp: 'cmm 8/22/2002 19:25'!
filename

	^memberIndex filename! !

!MaLargeCollectionManager methodsFor: 'accessing' stamp: 'cmm 3/17/2005 22:42'!
filenames

	^ { self filename }! !

!MaLargeCollectionManager methodsFor: 'testing' stamp: 'cmm 2/25/2003 10:19'!
isOpen

	^memberIndex ~= nil! !

!MaLargeCollectionManager methodsFor: 'accessing' stamp: ''!
memberIndex

	^memberIndex
! !

!MaLargeCollectionManager methodsFor: 'override' stamp: 'cmm 3/17/2005 23:14'!
openIndex: filenameString

	self subclassResponsibility! !

!MaLargeCollectionManager methodsFor: 'override' stamp: 'cmm 3/16/2005 22:57'!
process: aMaCollectionChanges recordResultsIn: aMaCommitResult

	self subclassResponsibility! !

!MaLargeCollectionManager methodsFor: 'override' stamp: 'cmm 3/16/2005 14:17'!
processAddsIn: aMaCollectionChanges using: aMaCommitResult

	self subclassResponsibility! !

!MaLargeCollectionManager methodsFor: 'override' stamp: 'cmm 3/16/2005 21:24'!
recordConflictsBetween: aMagmaCollectionChanges and: challengingCommitLogEntry in: myCommitLogEntry
	"If there are any conflicts, record them in myCommitLogEntry."

	self subclassResponsibility! !

!MaLargeCollectionManager methodsFor: 'override' stamp: 'cmm 3/16/2005 21:24'!
recordGeneralConflictsWith: aMagmaLargeCollectionChanges in: aMaCommitLogEntry

	self subclassResponsibility! !

!MaLargeCollectionManager methodsFor: 'private' stamp: ''!
repository: aMaObjectRepository

	repository := aMaObjectRepository
! !

!MaLargeCollectionManager methodsFor: 'accessing' stamp: ''!
size

	^memberIndex size
! !

!MaLargeCollectionManager methodsFor: 'override' stamp: 'cmm 4/30/2005 11:16'!
upTo: anInteger inSegmentFromIndex: lowIndex

	self subclassResponsibility! !

MaLargeCollectionManager subclass: #MagmaArrayManager
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-private'!

!MagmaArrayManager methodsFor: 'initializing' stamp: 'cmm 4/28/2005 20:40'!
openIndex: filenameString

	^ MaLargeArrayOfNumbers open: filenameString! !

!MagmaArrayManager methodsFor: 'actions' stamp: 'cmm 3/18/2005 10:55'!
process: aMaCollectionChanges recordResultsIn: aMaCommitResult

	self
		processAddsIn: aMaCollectionChanges
		using: aMaCommitResult.! !

!MagmaArrayManager methodsFor: 'private' stamp: 'cmm 3/18/2005 10:58'!
processAddsIn: aMaCollectionChanges using: aMaCommitResult

	aMaCollectionChanges additionsDo:
		[ : eachIndex : eachOid |  | addedObjectOid |
		addedObjectOid :=
			(MaOidCalculator isOidForNewObject: eachOid)
				ifTrue: [ aMaCommitResult permanentOids at: eachOid ]
				ifFalse: [ eachOid ].
		memberIndex
			at: eachIndex
			put: addedObjectOid ]! !

!MagmaArrayManager methodsFor: 'overriding' stamp: 'cmm 3/20/2005 21:29'!
recordConflictsBetween: aMagmaArrayChanges and: challengingCommitLogEntry in: myCommitLogEntry
	"For MagmaArray's, different sessions may update the Array as long as they update different indexes."

	(challengingCommitLogEntry hasChangesToMagmaArray: aMagmaArrayChanges collectionOid)
		ifTrue:
			[ | foreignChanges |
			foreignChanges _ challengingCommitLogEntry magmaArrayAdditionsAt: aMagmaArrayChanges collectionOid.
			aMagmaArrayChanges additionsDo:
				[ : eachIndex : eachOid |
				((foreignChanges includesKey: eachIndex)
				and: [ (foreignChanges at: eachIndex) ~= eachOid ])
					ifTrue:
						[ "this user tried to put a different oid at the same index, conflict!!"
						myCommitLogEntry
							failResultWith: challengingCommitLogEntry connection
							on: aMagmaArrayChanges collectionOid ] ] ]! !

!MagmaArrayManager methodsFor: 'overriding' stamp: 'cmm 3/18/2005 13:33'!
recordGeneralConflictsWith: aMagmaLargeCollectionChanges in: aMaCommitLogEntry

	"no general conflicts for MagmaArray's."! !

!MagmaArrayManager methodsFor: 'overriding' stamp: 'cmm 5/2/2005 22:19'!
upTo: anInteger inSegmentFromIndex: lowIndex

	| oids |
	oids _ 
		memberIndex
			upTo: anInteger
			startingAt: lowIndex.
	^ MagmaLargeCollectionSegment new
		lastKnownSize: memberIndex size ;
		startIndex: lowIndex ;
		objects: oids ;
		yourself! !

MaLargeCollectionManager subclass: #MagmaCollectionManager
	instanceVariableNames: 'indexesByAttribute indexes oidOfIndexesCollection lockedBy'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-private'!

!MagmaCollectionManager methodsFor: 'locking' stamp: 'cmm 3/16/2005 14:13'!
beLockedBy: aMagmaClientConnection

	lockedBy _ aMagmaClientConnection! !

!MagmaCollectionManager methodsFor: 'initialize-release' stamp: 'cmm 3/16/2005 14:24'!
close

	super close.
	indexes do: [ :each | each close ].
	self initializeIndexesByAttribute! !

!MagmaCollectionManager methodsFor: 'initialize-release' stamp: 'cmm 4/29/2005 12:07'!
createHashIndex: aMaIndexDefinition

	(MaHashIndex
		create: (self fullFileNameStringFor: aMaIndexDefinition)
		keySize: aMaIndexDefinition keySize
		valueSize: MaObjectBuffer oidSize
		recordSize: aMaIndexDefinition recordSize) close! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 4/28/2005 22:19'!
dummyValue
	"For memberIndex, only the keys matter.  What members do I have?  Don't need the values to know that.  This is the value that goes in all keys of the memberIndex."

	^ 0! !

!MagmaCollectionManager methodsFor: 'initialize-release' stamp: 'cmm 3/17/2005 22:48'!
ensureIndex: aMaIndexDefinition

	(repository directory fileExists: (self filenameFor: aMaIndexDefinition))
		ifFalse: 
			[ self createHashIndex: aMaIndexDefinition ]! !

!MagmaCollectionManager methodsFor: 'initialize-release' stamp: 'cmm 3/16/2005 14:14'!
ensureIndexesOpen

	collection indexesDo:
		[ :each |
		(self hasIndexFor: each attribute)
			ifFalse:
				[ indexes add:
					(indexesByAttribute
						at: each attribute
						put: (self openIndex: (self fullFileNameStringFor: each))) ] ]! !

!MagmaCollectionManager methodsFor: 'actions' stamp: 'cmm 3/17/2005 17:22'!
ensureOpen

	super ensureOpen.
	self ensureIndexesOpen! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 3/16/2005 21:28'!
filenameComponentFor: aMaIndexDefinition

	^ self collectionOid printString , aMaIndexDefinition uniqueFileIdentifier
! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 3/20/2005 19:20'!
filenameFor: aMaIndexDefinition

	^ (String streamContents:
		[ : stream |
		stream
			nextPutAll: (self filenameComponentFor: aMaIndexDefinition) ;
			nextPut: $. ;
			nextPutAll: collection changes suggestedFileExtension ])! !

!MagmaCollectionManager methodsFor: 'accessing' stamp: 'cmm 3/17/2005 22:42'!
filenames

	^ super filenames, (indexes collect: [ : each | each filename ])! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 3/17/2005 22:47'!
fullFileNameStringFor: aMaIndexDefinition

	^ repository directory fullNameFor: (self filenameFor: aMaIndexDefinition)! !

!MagmaCollectionManager methodsFor: 'testing' stamp: 'cmm 3/16/2005 14:15'!
hasIndexFor: attributeSymbol

	^ indexesByAttribute includesKey: attributeSymbol! !

!MagmaCollectionManager methodsFor: 'accessing' stamp: 'cmm 3/18/2005 14:25'!
indexForAttribute: aSymbol

	^self
		indexForAttribute: aSymbol
		ifNone:
			[ MagmaEnvironmentError signal:
				'The index for ' , aSymbol ,
					' was not found.  (MaLargeCollectionManager>>#indexForAttribute:).' ]
! !

!MagmaCollectionManager methodsFor: 'accessing' stamp: 'cmm 3/16/2005 14:14'!
indexForAttribute: aSymbol ifNone: aBlock

	aSymbol isNil ifTrue: [ ^self memberIndex ].
	^indexesByAttribute
		at: aSymbol
		ifAbsent: aBlock
! !

!MagmaCollectionManager methodsFor: 'initialize-release' stamp: 'cmm 3/16/2005 14:19'!
initialize

	super initialize.
	"we need them mapped to the correct attribute."
	self initializeIndexesByAttribute.
	"and we must know the correct order, as well"
	indexes _ OrderedCollection new! !

!MagmaCollectionManager methodsFor: 'initialize-release' stamp: 'cmm 3/16/2005 14:15'!
initializeIndexesByAttribute

	indexesByAttribute := Dictionary new! !

!MagmaCollectionManager methodsFor: 'testing' stamp: 'cmm 3/16/2005 14:22'!
isLocked

	^lockedBy ~= nil! !

!MagmaCollectionManager methodsFor: 'accessing' stamp: 'cmm 3/16/2005 14:22'!
lockedBy
	"Answers a MagmaClientConnection."

	^lockedBy! !

!MagmaCollectionManager methodsFor: 'accessing' stamp: 'cmm 3/16/2005 14:13'!
oidOfIndexesCollection: anInteger

	oidOfIndexesCollection := anInteger
! !

!MagmaCollectionManager methodsFor: 'initialize-release' stamp: 'cmm 3/17/2005 22:35'!
openIndex: aString

	^MaHashIndex open: aString
! !

!MagmaCollectionManager methodsFor: 'overriding' stamp: 'cmm 3/16/2005 22:57'!
process: aMaCollectionChanges recordResultsIn: aMaCommitResult

	self
		processLockIn: aMaCollectionChanges
		using: aMaCommitResult connection.
	self processRemovedIndexesIn: aMaCollectionChanges.
	self
		processAddsIn: aMaCollectionChanges
		using: aMaCommitResult.
	self
		processRemovesIn: aMaCollectionChanges;
		processChangedIndexValuesIn: aMaCollectionChanges;
		processAddNewIndexValuesIn: aMaCollectionChanges! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 3/16/2005 22:58'!
processAddNewIndexValuesIn: aMaCollectionChanges
	"Serialize the new values for an index that is being built."

	aMaCollectionChanges hasNewIndexValues ifFalse: [ ^self ].
	aMaCollectionChanges newIndexValues keysAndValuesDo:
		[ :eachAttribute :eachValues | | index |
		index := (self indexForAttribute: eachAttribute).
		eachValues do:
			[ :eachAssociation |
			eachAssociation key do:
				[ :eachKey |
				index
					add: eachAssociation value
					at: eachKey ] ] ]! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 4/28/2005 22:19'!
processAddsIn: aMaCollectionChanges using: aMaCommitResult

	aMaCollectionChanges additionsDo:
		[ : eachOid : eachHashIndexArrays | | addedObjectOid |
		addedObjectOid :=
			(MaOidCalculator isOidForNewObject: eachOid)
				ifTrue: [ aMaCommitResult permanentOids at: eachOid ]
				ifFalse: [ eachOid ].
		memberIndex
			add: self dummyValue
			at: addedObjectOid.
		eachHashIndexArrays
			with: indexes
			do:
				[ : eachHashValues : eachIndex |
				eachHashValues do:
					[ :eachHashValue |
					eachIndex
						add: addedObjectOid
						at: eachHashValue ] ] ]! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 4/6/2005 21:41'!
processChangedIndexValuesIn: aMaCollectionChanges

	aMaCollectionChanges changedKeySpecificationsDo:
		[ :each | | index |
		index := self indexForAttribute: each attribute.
		"Only remove has values unique to each collection.  Don't waste processing by removing and then adding at the same key."
		(memberIndex includesKey: each object)
			ifTrue:
				[ each oldHashValues do:
					[ :eachHashKeyValue | 
					index
						remove: each oid
						at: eachHashKeyValue ].
				each newHashValues do:
					[ :eachHashKeyValue |
					index
						add: each oid
						at: eachHashKeyValue ] ] ]! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 3/16/2005 22:58'!
processLockIn: aMaCollectionChanges using: aMagmaConnection

	aMaCollectionChanges hasLockSpecified
		ifTrue:
			[ aMaCollectionChanges isLocked
				ifTrue: [ self beLockedBy: aMagmaConnection ]
				ifFalse: [ self beLockedBy: nil ] ]! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 3/16/2005 22:58'!
processRemovedIndexesIn: aMaCollectionChanges

	aMaCollectionChanges
		removedIndexesDo: [ :each | self removeIndex: (self indexForAttribute: each) ]
! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 4/28/2005 22:19'!
processRemovesIn: aMaLargeCollectionChanges

	aMaLargeCollectionChanges removedDo:
		[ :eachOid :eachHashIndexValuesArray |
		memberIndex
			remove: self dummyValue
			at: eachOid.
		eachHashIndexValuesArray
			with: indexes
			do:
				[ :eachArray :eachIndex |
				eachArray do:
					[ :eachKeyValue |
					eachIndex
						remove: eachOid
						at: eachKeyValue ] ] ]! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 3/16/2005 21:20'!
recordConflictsBetween: aMagmaCollectionChanges and: challengingCommitLogEntry in: myCommitLogEntry

	"Fail my result if any indexes have been added or removed since my last transaction boundary."
	(challengingCommitLogEntry hasCommitted: oidOfIndexesCollection)
		ifTrue:
			[ myCommitLogEntry
				failResultWith: challengingCommitLogEntry connection
				on: oidOfIndexesCollection ].
	"Fail my result if any objects being added or removed have been changed by any other session."
	aMagmaCollectionChanges additionsDo:
		[ : eachOid : eachHashIndexValues |
		(challengingCommitLogEntry hasCommitted: eachOid)
			ifTrue:
				[ myCommitLogEntry
					failResultWith: challengingCommitLogEntry connection
					on: eachOid ] ].
	aMagmaCollectionChanges removedDo:
		[ : eachOid : eachHashIndexValues |
		(challengingCommitLogEntry hasCommitted: eachOid)
			ifTrue:
				[ myCommitLogEntry
					failResultWith: challengingCommitLogEntry connection
					on: eachOid ] ]! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 3/16/2005 21:09'!
recordGeneralConflictsWith: aMagmaLargeCollectionChanges in: aMaCommitLogEntry

	"If its locked by someone else, no changes are allowed."
	(self isLocked
	and: [ aMaCommitLogEntry connection ~= lockedBy ])
		ifTrue:
			[ aMaCommitLogEntry
				failResultWith: self lockedBy
				on: aMagmaLargeCollectionChanges collectionOid ]! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 3/16/2005 22:58'!
removeIndex: aMaHashIndex

	aMaHashIndex
		close;
		delete
! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 4/17/2005 22:13'!
upTo: anInteger inSegmentFrom: lowKey to: highKey startingAt: indexInteger on: attributeSymbol

	| index keysAndObjects |
	index _ self indexForAttribute: attributeSymbol.
	keysAndObjects _
		index
			upTo: anInteger
			keysAndValuesFrom: lowKey
			to: highKey
			startingAt: indexInteger.
	^ MagmaLargeCollectionSegment new
		lastKnownSize:
			(index
				numberOfEntriesFrom: lowKey
				to: highKey) ;
		keys: keysAndObjects first ;
		objects: keysAndObjects last ;
		startIndex: indexInteger ;
		yourself! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 5/2/2005 22:19'!
upTo: anInteger inSegmentFromIndex: lowIndex

	| keysAndValues |
	keysAndValues _  
		memberIndex
			upTo: anInteger
			keysAndValuesFromIndex: lowIndex.
	^ MagmaLargeCollectionSegment new
		lastKnownSize: memberIndex size ;
		startIndex: lowIndex ;
		objects: keysAndValues first  "there are no 'values' in the member-index, the keys are the objects." ;
		yourself! !

MaObject subclass: #MaObjectFiler
	instanceVariableNames: 'directory file preMadeObjectBuffer filePositionIndex usedByteArrays primitiveAttributeAddressesMap'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-private'!

!MaObjectFiler class methodsFor: 'creation' stamp: 'cmm 9/24/2004 12:52'!
create: aFileDirectory

	^ self new create: aFileDirectory! !

!MaObjectFiler class methodsFor: 'accessing' stamp: 'cmm 4/28/2005 11:36'!
maxDbSize
	"Answers the maximum size of a Magma database as a power of 2.  Hmm, wow, 281 Terabytes huh..  This seems to be beyond any practical limits for a single repository; at least in year 2005, so we'll go with this for now."

	^ 48! !

!MaObjectFiler class methodsFor: 'accessing' stamp: 'cmm 9/24/2004 13:01'!
objectPointersFilename

	^ 'objects.idx'! !

!MaObjectFiler class methodsFor: 'accessing' stamp: 'cmm 9/24/2004 13:01'!
objectsFilename

	^ 'objects.magma'! !

!MaObjectFiler class methodsFor: 'creation' stamp: 'cmm 9/24/2004 13:10'!
open: aFileDirectory

	^ self new
		directory: aFileDirectory ;
		open! !

!MaObjectFiler class methodsFor: 'accessing' stamp: ''!
sizeOfEmptyRepository
	"Answer in bytes."

	^1024
! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'cmm 12/2/2004 13:38'!
anchorOid

	^ self primitiveAttributeAt: 'anchorOid'! !

!MaObjectFiler methodsFor: 'building' stamp: 'cmm 12/2/2004 13:38'!
anchorOid: anInteger

	^ self
		primitiveAttributeAt: 'anchorOid'
		put: anInteger! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 11/24/2004 15:33'!
append: aMaObjectBuffer to: aMaSerializedGraphBuffer withDataAt: filePositionInteger

	| startPos |

	startPos _ aMaSerializedGraphBuffer bufferSize.
	self
		ensureByteArrayIn: aMaObjectBuffer
		and: aMaSerializedGraphBuffer
		canAccomodate: self trackSize + aMaObjectBuffer bufferPositionOfBytesToStore
		bytesFrom: startPos + 1.
	aMaSerializedGraphBuffer increaseBufferSizeBy:
		(self
			read: self trackSize
			bytesInto: aMaObjectBuffer
			and: aMaSerializedGraphBuffer
			startingAt: startPos
			filePosition: filePositionInteger)! !

!MaObjectFiler methodsFor: 'actions' stamp: 'cmm 1/25/2005 14:10'!
appendObject: oidInteger into: aMaSerializedGraphBuffer
	"Read a single object into aMaSerializedGraphBuffer and adjust its bufferSize accordingly."

	| buffer |
	preMadeObjectBuffer
		byteArray: aMaSerializedGraphBuffer byteArray;
		startPos: aMaSerializedGraphBuffer bufferSize.
	self
		append: preMadeObjectBuffer
		to: aMaSerializedGraphBuffer
		withDataAt: (self filePointerForOid: oidInteger).
	buffer _ preMadeObjectBuffer asSentientBuffer.
	"Set oid only after append because that is where the need for growth is checked."
	buffer oid: oidInteger.
	^buffer
! !

!MaObjectFiler methodsFor: 'debugging' stamp: 'cmm 5/3/2005 14:26'!
bufferFor: oidInteger

	| buffer |
	buffer _ MaSerializedGraphBuffer new: 1000.
	self
		appendObject: oidInteger
		into: buffer.
	^ buffer rootBuffer! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'cmm 12/2/2004 13:40'!
classDefinitionsOid

	^self primitiveAttributeAt: 'classDefinitionsOid'
! !

!MaObjectFiler methodsFor: 'building' stamp: 'cmm 12/2/2004 13:40'!
classDefinitionsOid: anInteger

	self
		primitiveAttributeAt: 'classDefinitionsOid'
		put: anInteger
! !

!MaObjectFiler methodsFor: 'actions' stamp: 'cmm 9/24/2004 13:22'!
close

	file ifNotNil:
		[ file closed ifFalse: [ file close ].
		file _ nil ].
	filePositionIndex ifNotNil:
		[ filePositionIndex close.
		filePositionIndex _ nil ]! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 9/24/2004 16:42'!
create: aFileDirectory

	self
		directory: aFileDirectory assureExistence ;
		createDataFile ;
		createFilePositionIndex.
	^ self! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 3/11/2005 14:40'!
createDataFile

	(directory fileExists: self class objectsFilename)
		ifTrue:
			[ MagmaEnvironmentError signal: self dataFileName, ' already exists.' ].
	self openObjectsFile.
	file
		next: self class sizeOfEmptyRepository
		putAll: (ByteArray new: self class sizeOfEmptyRepository)
		startingAt: 1.
	self
		classDefinitionsOid: MaOidCalculator oidForNil;
		definitionOid: MaOidCalculator oidForNil;
		signature: MaObjectRepository recognitionSignature;
		version: MagmaSession magmaVersion! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 5/2/2005 21:23'!
createFilePositionIndex

	filePositionIndex _
		MaLargeArrayOfNumbers
			create: self filePositionIndexFileName
			bitSize: self class maxDbSize.
	filePositionIndex open: self filePositionIndexFileName! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'cmm 9/24/2004 13:02'!
dataFileName

	^ directory fullNameFor: self class objectsFilename! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'cmm 12/2/2004 13:42'!
definitionOid

	^ self primitiveAttributeAt: 'definitionOid'! !

!MaObjectFiler methodsFor: 'building' stamp: 'cmm 12/2/2004 13:42'!
definitionOid: anInteger

	^self
		primitiveAttributeAt: 'definitionOid'
		put: anInteger
! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'cmm 9/26/2004 20:25'!
directory
	^ directory! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 9/24/2004 12:55'!
directory: aFileDirectory

	directory _ aFileDirectory! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 12/30/2002 17:28'!
ensureByteArrayIn: aMaObjectBuffer and: aMaSerializedGraphBuffer canAccomodate: anInteger bytesFrom: startPos

	(startPos + anInteger) > aMaObjectBuffer byteArray size
		ifTrue:
			[ | bigByteArray beforeSize afterSize |
			beforeSize _ aMaObjectBuffer byteArray size.
			afterSize _ (aMaObjectBuffer byteArray size * 2) max: startPos + anInteger.
			bigByteArray _ ByteArray new: afterSize.
			bigByteArray
				replaceFrom: 1
				to: beforeSize
				with: aMaSerializedGraphBuffer byteArray
				startingAt: 1.
			aMaObjectBuffer byteArray: bigByteArray.
			aMaSerializedGraphBuffer byteArray: bigByteArray ]! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'cmm 3/1/2005 17:36'!
filePointerForOid: oidInteger

	| answer |
	answer _ filePositionIndex 
		at: (MaOidCalculator indexPositionForOid: oidInteger)
		ifAbsent: [ MagmaInvalidOid signal: oidInteger printString, ' is not a valid oid.' ].
	answer = 0 ifTrue: [ MagmaInvalidOid signal: oidInteger printString, ' is not a valid oid.' ].
	^ answer! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'cmm 9/24/2004 13:07'!
filePositionIndexFileName

	^ directory fullNameFor: self class objectPointersFilename! !

!MaObjectFiler methodsFor: 'actions' stamp: ''!
finalize

	self close
! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'cmm 12/2/2004 13:32'!
getNextOid

	^MaOidCalculator oidForFilePositionIndexPosition: filePositionIndex size + 1! !

!MaObjectFiler methodsFor: 'actions' stamp: 'cmm 12/2/2004 13:45'!
index: oidInteger at: filePositionInteger

	^filePositionIndex
		at: (MaOidCalculator indexPositionForOid: oidInteger)
		put: filePositionInteger! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 10/21/2004 12:53'!
initialize

	super initialize.
	preMadeObjectBuffer _ MaObjectBuffer new.
	usedByteArrays _ (1 to: 8) collect: [ :each | ByteArray new: each ].
	self initializePrimitiveAttributeAddressesMap! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 12/29/2004 00:22'!
initializePrimitiveAttributeAddressesMap

	primitiveAttributeAddressesMap _ Dictionary new
		at: 'signature' put: (Association key: 0 value: 64);
		at: 0 put: 'signature';

		at: 'version' put: (Association key: 8 value: 16);
		at: 2 put: 'version';

		at: 'booleanFlags' put: (Association key: 10 value: 8);
		at: 10 put: 'booleanFlags';

		at: 'definitionOid' put: (Association key: 11 value: 64);
		at: 11 put: 'definitionOid';

		"gap 19-26"
		at: 'classDefinitionsOid' put: (Association key: 27 value: 64);
		at: 27 put: 'classDefinitionsOid';

		at: 'anchorOid' put: (Association key: 43 value: 64);
		at: 43 put: 'anchorOid';

		yourself! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'cmm 5/1/2003 17:37'!
isOpen

	^file notNil and: [ file closed not ]! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'cmm 4/21/2003 16:29'!
objectFileSize

	^file size! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'cmm 4/28/2005 20:40'!
open

	(directory fileExists: self class objectsFilename)
		ifFalse:
			[ MagmaEnvironmentError signal: self dataFileName, ' not found.' ].
	self openObjectsFile.
	[ self validateSignature ] ifCurtailed: [ self close ].
	[ self validateVersion ] ifCurtailed: [ self close ].
	filePositionIndex _
		[ MaLargeArrayOfNumbers open: self filePositionIndexFileName ] ifCurtailed: [ self close ].
	^ self! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 3/11/2005 14:41'!
openObjectsFile

	file _ directory fileNamed: self class objectsFilename.
	file
		ifNil: [ MagmaEnvironmentError signal: 'Could not open files in ', directory pathName ]
		ifNotNil: [ file binary ]! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'cmm 10/21/2004 23:00'!
physicalSizeOfObjectAt: filePointer
	"Used to determine how much space we can deallocate from my 'memory' because an object, most likely an OrderedCollection, residing at filePointer has grown and needs to be relocated."

	| byteArray buffer |

	byteArray _ ByteArray new: MaObjectBuffer headerSize.
	buffer _
		MaObjectBuffer
			byteArray: byteArray
			startPos: 0.
	file
		maRead: byteArray size - buffer bufferPositionOfBytesToStore + 1
		bytesFromPosition: buffer bufferPositionOfBytesToStore
		of: byteArray
		atFilePosition: filePointer.
	^ buffer physicalSize! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 12/30/2002 21:15'!
primitiveAttributeAddressesMap

	^primitiveAttributeAddressesMap! !

!MaObjectFiler methodsFor: 'accessing' stamp: ''!
primitiveAttributeAt: aString

	| assoc |

	^self
		uint: (assoc _ self primitiveAttributeAddressesMap at: aString) value
		at: assoc key
! !

!MaObjectFiler methodsFor: 'private' stamp: ''!
primitiveAttributeAt: aString put: anInteger

	| assoc |

	self
		uint: (assoc _ self primitiveAttributeAddressesMap at: aString) value
		at: assoc key
		put: anInteger
! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 9/24/2004 13:21'!
read: numberOfBytes bytesInto: aMaObjectBuffer and: aMaSerializedGraphBuffer startingAt: startPos filePosition: anInteger
	"startPos is 0-based, answer the physicalSize of the buffer read."

	| physicalSize |

	file
		maRead: numberOfBytes
		bytesFromPosition: startPos + aMaObjectBuffer bufferPositionOfBytesToStore
		of: aMaObjectBuffer byteArray
		atFilePosition: anInteger.
	"See if we have read enough, if not read the rest of the buffer."
	(physicalSize _ aMaObjectBuffer physicalSize) > numberOfBytes
		ifTrue:
			[ self
				ensureByteArrayIn: aMaObjectBuffer
				and: aMaSerializedGraphBuffer
				canAccomodate: physicalSize
				bytesFrom: startPos + aMaObjectBuffer bufferPositionOfBytesToStore.
			file
				next: physicalSize - numberOfBytes
				into: aMaSerializedGraphBuffer byteArray
				startingAt: startPos + numberOfBytes + aMaObjectBuffer bufferPositionOfBytesToStore ].
	^physicalSize! !

!MaObjectFiler methodsFor: 'accessing' stamp: ''!
signature

	^self primitiveAttributeAt: 'signature'
! !

!MaObjectFiler methodsFor: 'private' stamp: ''!
signature: anInteger

	self
		primitiveAttributeAt: 'signature'
		put: anInteger
! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 8/12/2002 23:48'!
trackSize
	"This is the minimum number of bytes to read from disk for any object.
	This particular value allows up to 32 instVars to be read at once."

	"^MaAbstractObjectBuffer addressSize * 32 + MaNormalObjectBuffer headerSize"
	^280 "same as above, but faster.."! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 9/24/2004 13:22'!
uint: bits at: anInteger

	| address bytes |

	address _ self usedByteArray: (bytes _ bits / 8).
	file
		maRead: bytes
		bytesFromPosition: 1
		of: address
		atFilePosition: anInteger.
	^address
		maUint: bits
		at: 0
! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 9/24/2004 13:22'!
uint: bits at: filePointer put: anInteger

	| ba bytes |

	ba _ self usedByteArray: (bytes _ bits / 8).
	ba
		maUint: bits
		at: 0
		put: anInteger.
	file
		maWrite: bytes
		bytesFromPosition: 1
		of: ba
		atFilePosition: filePointer
! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 7/16/2002 00:08'!
usedByteArray: size

	^usedByteArrays at: size
! !

!MaObjectFiler methodsFor: 'actions' stamp: 'cmm 7/16/2002 00:01'!
validateSignature

	self signature = MaObjectRepository recognitionSignature
		ifFalse:
			[ MagmaUserError signal: 'This does not appear to be a Magma database.' ]
! !

!MaObjectFiler methodsFor: 'actions' stamp: 'cmm 7/16/2002 21:55'!
validateVersion

	self version = MagmaSession magmaVersion
		ifFalse:
			[ MagmaUserError signal: 'This database requires code version ', self version printString, '.  You are running version ', MagmaSession magmaVersion printString ]! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'cmm 7/16/2002 21:53'!
version

	^self primitiveAttributeAt: 'version'! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 7/16/2002 21:53'!
version: anInteger

	self
		primitiveAttributeAt: 'version'
		put: anInteger
! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 1/19/2005 15:40'!
write: aMaObjectBuffer at: filePosition
	"Note that the oid is stored in the filePositionsIndex, so there is no need to store it in the data file too.  The oid is the first 8-bytes of the buffer."

	file
		maWrite: aMaObjectBuffer numberOfBytesToStore
		bytesFromPosition:
			aMaObjectBuffer startPos + aMaObjectBuffer bufferPositionOfBytesToStore
		of: aMaObjectBuffer byteArray
		atFilePosition: filePosition! !

!MaObjectFiler methodsFor: 'actions' stamp: 'cmm 12/2/2004 13:45'!
write: aMaObjectBuffer index: aBoolean
	"write a single object encapsulated by aMaObjectBuffer."

	| filePosition |
	filePosition _
		aBoolean
			ifTrue:
				[ self
					index: aMaObjectBuffer oid
					at: file size ]
			ifFalse:
				[ | pos |
				pos _ self filePointerForOid: aMaObjectBuffer oid.
				(self physicalSizeOfObjectAt: pos) < aMaObjectBuffer physicalSize
					ifTrue: [ self index: aMaObjectBuffer oid at: file size ]
					ifFalse: [ pos ] ].
	self
		write: aMaObjectBuffer
		at: filePosition! !

MaObject subclass: #MaObjectRepository
	instanceVariableNames: 'transactionLog sessions filer repositoryController largeCollectionManagers systemReadStrategy nextOid'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-private'!

!MaObjectRepository class methodsFor: 'creation' stamp: 'cmm 9/24/2004 12:50'!
create: aFileDirectory controller: aMaRepositoryController

	^ (self new repositoryController: aMaRepositoryController) create: aFileDirectory! !

!MaObjectRepository class methodsFor: 'creation' stamp: 'cmm 9/24/2004 15:39'!
open: aFileDirectory controller: aMaRepositoryController

	^self new
		repositoryController: aMaRepositoryController;
		open: aFileDirectory! !

!MaObjectRepository class methodsFor: 'accessing' stamp: ''!
recognitionSignature

	^5191969
! !

!MaObjectRepository methodsFor: 'actions' stamp: 'cmm 1/18/2005 22:39'!
abortTransactionFor: sessionId

	^ transactionLog 
		abortTransactionEntryFor: sessionId
		repository: self! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 2/8/2005 23:37'!
anchorByteArrayUsing: aMaReadStrategy

	| anchorOid |
	^ MaOidCalculator oidForNil = (anchorOid _ filer anchorOid)
		ifTrue:
			[ (self serverSerializer serializeGraph: MaRootAnchor new) trimmedByteArray ]
		ifFalse:
			[ self
				byteArrayAt: anchorOid
				using: aMaReadStrategy ]! !

!MaObjectRepository methodsFor: 'actions' stamp: 'cmm 3/16/2005 21:33'!
assignPermanentOidFor: aMaObjectBuffer mapUsing: aDictionary andAddNewObjectOidTo: aSet
	"aMaObjectBuffer is going to written very soon, however, none of its oids are permanent.  This method assigns and populates aDictionary with a map of the temporary to the permanent oids for aMaObjectBuffer (not the oids they link, however) as it populates its oid."

	| newOid |
	aMaObjectBuffer oid: (
		aDictionary
			at: aMaObjectBuffer oid
			put: (newOid _ self getNextOid)).
	aSet add: newOid! !

!MaObjectRepository methodsFor: 'private' stamp: 'cmm 12/2/2004 13:27'!
assignPermanentOidsIn: aMaCommitPackage andAddNewObjectOidsTo: aSet

	| tempToPermOidsMap |

	tempToPermOidsMap _ Dictionary new: 50.
	aMaCommitPackage newObjectsDo:
		[ :eachBuffer |
		self
			assignPermanentOidFor: eachBuffer
			mapUsing: tempToPermOidsMap
			andAddNewObjectOidTo: aSet ].
	^ tempToPermOidsMap! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 3/19/2005 10:59'!
byteArrayAt: oidInteger using: aMaReadStrategy

	^ (MaOidCalculator isOidForUserObject: oidInteger)
		ifTrue:
			[ (self
				graphBufferAt: oidInteger
				using: aMaReadStrategy) trimmedByteArray ]
		ifFalse:
			[ "When could this happen?  (Sigh), when they put an atomic in a MagmaLargeCollection."
			(self session serializer serializeGraph:
				(MaOidCalculator
					objectWithOid: oidInteger
					ifNone: [ MagmaSoftwareError signal: 'unknown oid' ])) trimmedByteArray ]! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 12/2/2004 13:41'!
classDefinitionsByteArray
	"When initially creating the repository, before even the system root is
	written, the classDefinitionsOid will be a new-object oid."

	| loid |

	^MaOidCalculator oidForNil = (loid _ filer classDefinitionsOid)
		ifTrue:
			[ (self serverSerializer serializeGraph: repositoryController definition classDefinitions) trimmedByteArray ]
		ifFalse:
			[ self
				byteArrayAt: loid
				using: MaBasicReadStrategy deep "we need it all" ]! !

!MaObjectRepository methodsFor: 'actions' stamp: 'cmm 12/18/2002 22:26'!
close

	filer ifNotNil: [ filer close ].
	largeCollectionManagers do: [ :each | each close ]! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 3/19/2005 11:08'!
collectionFilenameFor: oidInteger extension: aString

	^ self directory fullNameFor:
		(String streamContents:
			[ : stream |
			stream
				print: oidInteger ;
				nextPut: $. ;
				nextPutAll: aString ])! !

!MaObjectRepository methodsFor: 'magma-collections' stamp: 'cmm 7/2/2004 16:26'!
collectionManagerFor: oidInteger

	^ largeCollectionManagers
		at: oidInteger
		ifAbsent:
			[ MagmaSoftwareError signal:
				'MagmaCollection ' , oidInteger printString ,
					' not found.  MaObjectRepository>>collectionManagerFor:' ]! !

!MaObjectRepository methodsFor: 'private' stamp: 'cmm 1/18/2005 22:30'!
commitEntryFor: sessionId beginAnother: aBoolean

	^ transactionLog
		renewEntryFor: sessionId
		beginAnother: aBoolean
		for: self! !

!MaObjectRepository methodsFor: 'private' stamp: 'cmm 12/2/2004 13:26'!
create: aFileDirectory

	filer _ MaObjectFiler create: aFileDirectory.
	nextOid _ filer getNextOid.
	^ self! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 12/2/2004 13:43'!
definitionByteArray

	| oid |
	^MaOidCalculator oidForNil = (oid _ filer definitionOid)
		ifTrue:
			[ (self serverSerializer serializeGraph: repositoryController definition) trimmedByteArray ]
		ifFalse:
			[ self
				byteArrayAt: oid
				using: MaBasicReadStrategy deep "we need it all" ]! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 9/24/2004 15:42'!
directory

	^ filer directory! !

!MaObjectRepository methodsFor: 'building' stamp: 'cmm 3/16/2005 23:08'!
ensureManagerCreatedFor: aMagmaLargeCollection

	| answer |
	answer _ 
		largeCollectionManagers
			at: (self session oidFor: aMagmaLargeCollection)
			ifAbsentPut: [ aMagmaLargeCollection newManagerUsing: self ].
	^ answer! !

!MaObjectRepository methodsFor: 'private' stamp: 'cmm 3/11/2005 15:14'!
establishLinksToNewObjectsIn: aMaCommitPackage using: newOidToPermanentOidMap
	"Update all new-object oids in aMaCommitPackage
	with their permanent counter-part.  newOidToPermanentOidMap has an entry for
	them all."

	aMaCommitPackage objectsDo:
		[ :eachBuffer |
		eachBuffer isPointers
			ifTrue:
				[ eachBuffer instVarsDoWithIndex:
					[ :oid :index |
					(MaOidCalculator isOidForNewObject: oid)
						ifTrue:
							[ eachBuffer
								maInstVarAt: index
								put:
									(newOidToPermanentOidMap 
										at: oid
										ifAbsent: [ MagmaUserError signal: 'Invalid reference, an object may not change during serialization.' ]) ] ] ] ]! !

!MaObjectRepository methodsFor: 'accessing' stamp: ''!
filer

	^filer
! !

!MaObjectRepository methodsFor: 'private' stamp: 'cmm 12/2/2004 13:25'!
getNextOid

	^nextOid _ nextOid + 1! !

!MaObjectRepository methodsFor: 'private' stamp: 'cmm 1/25/2005 14:04'!
graphBufferAt: oidInteger using: aMaReadStrategy
	"Answers a MaSerializedGraphBuffer."

	| answer rootBuffer |
	answer _ MaSerializedGraphBuffer new: 400.
	rootBuffer _ 
		filer
			appendObject: oidInteger
			into: answer.
	rootBuffer
		appendGraphUsing: aMaReadStrategy
		into: answer
		currentDepth: 1
		minDepth: aMaReadStrategy minimumDepth
		with: Set new
		filer: filer.
	^ answer! !

!MaObjectRepository methodsFor: 'private' stamp: 'cmm 4/28/2003 16:28'!
initialize

	super initialize.
	transactionLog _ MaTransactionLog new.
	largeCollectionManagers _ Dictionary new.
	sessions _ Dictionary new! !

!MaObjectRepository methodsFor: 'private' stamp: 'cmm 2/1/2005 22:45'!
initializeClassDefinitionsSpecialOids
	"This collection of special oids can grow.  When it does, clients need to know about it."

	repositoryController definition classDefinitions do:
		[ : each |
		(repositoryController session isPersistent: each)
			ifTrue:
				[ transactionLog
					onChangeOf: (repositoryController session oidFor: each "OrderedCollection of definitions")
					clientsShould: #refreshClassVersions ] ]! !

!MaObjectRepository methodsFor: 'private' stamp: 'cmm 2/1/2005 22:47'!
initializeSpecialOidsList
	"See MagmaSession>>#handleSpecialOidNotifications:"

	| classDefinitions session |
	classDefinitions _ repositoryController definition classDefinitions.
	session _ repositoryController session.
	(session isPersistent: classDefinitions)
		ifTrue:
			[ transactionLog
				onChangeOf: (repositoryController session oidFor: repositoryController definition classDefinitions)
				clientsShould: #refreshClassDefinitions ].
	self initializeClassDefinitionsSpecialOids! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 2/9/2005 10:26'!
initializeSystemReadStrategyUsing: aMaClassIdManager

	"The object and its immediate variables"
self maMarked: 'dev'.  "Should minimumDepth be 0?  If not, then perhaps we don't need to extend any Collection 1-level further, since the point for it was to merely get its immediately referenced objects...??"
	systemReadStrategy _ MaReadStrategy minimumDepth: 1.
	systemReadStrategy makeReadyForUseUsing: aMaClassIdManager! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 11/1/2004 10:23'!
isOpen

	^ filer notNil and: [ filer isOpen ]! !

!MaObjectRepository methodsFor: 'accessing' stamp: ''!
largeCollectionManagers

	^largeCollectionManagers! !

!MaObjectRepository methodsFor: 'private' stamp: 'cmm 12/2/2004 13:29'!
linkNewObjectsIn: aMaCommitPackage recordPermenantOidsUsing: aMaCommitResult
	"Some of the objects in each graphBuffers object buffers are new objects.  These need to be assigned a permanent loid and all of the objects that reference them need to have their references updated to reflect the new permanent loid."

	| newObjects tempToPermanentOidsMap |

	tempToPermanentOidsMap _
		self
			assignPermanentOidsIn: aMaCommitPackage
			andAddNewObjectOidsTo: (newObjects _ Set new: 100).
	self
		establishLinksToNewObjectsIn: aMaCommitPackage
		using: tempToPermanentOidsMap.
	aMaCommitResult permanentOids: tempToPermanentOidsMap.
	^ newObjects! !

!MaObjectRepository methodsFor: 'magma-collections' stamp: 'cmm 9/26/2004 21:06'!
magmaCollectionManagersDo: aBlock

	largeCollectionManagers do: aBlock! !

!MaObjectRepository methodsFor: 'building' stamp: 'cmm 6/7/2004 15:20'!
newSessionFor: userIdString numberOfChallengers: anInteger
	"Answer an Association whose key is the session id, value is the classMap
	array for that session."

	| connection |
	connection _
		transactionLog
			newSessionFor: userIdString
			numberOfChallengers: anInteger.
	sessions at: connection sessionId put: connection.
	^(Array new: 3)
		at: 1
			put: connection sessionId;
		at: 2
			put: self classDefinitionsByteArray ;
		at: 3
			put: self definitionByteArray;
		yourself! !

!MaObjectRepository methodsFor: 'building' stamp: 'cmm 1/28/2005 00:04'!
newTransactionFor: sessionIdInteger

	(transactionLog entryFor: sessionIdInteger) isCommitEntry
		ifTrue:
			[ MagmaSoftwareError signal: 'Already in a transaction.' ].
	^ transactionLog
		newTransactionFor: sessionIdInteger
		repository: self! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 12/2/2004 13:26'!
objectCount

	^nextOid - MaOidCalculator firstUserObjectOid! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 4/21/2003 16:29'!
objectFileSize

	^filer objectFileSize! !

!MaObjectRepository methodsFor: 'actions' stamp: 'cmm 12/2/2004 13:26'!
open: aFileDirectory

	filer _ MaObjectFiler open: aFileDirectory.
	nextOid _ filer getNextOid! !

!MaObjectRepository methodsFor: 'building' stamp: 'cmm 3/16/2005 23:06'!
openLargeCollectionIfNecessary: aMagmaLargeCollection

	(self ensureManagerCreatedFor: aMagmaLargeCollection) ensureOpen! !

!MaObjectRepository methodsFor: 'actions' stamp: 'cmm 7/2/2004 15:56'!
removeSessionFor: sessionId

	sessions
		removeKey: sessionId
		ifAbsent: [ MagmaUserError signal: 'session not connected.' ].
	transactionLog
		removeEntryForSessionId: sessionId
		ifAbsent: [ MagmaUserError signal: 'session not connected.' ]! !

!MaObjectRepository methodsFor: 'actions' stamp: 'cmm 4/28/2003 23:51'!
renewCurrentReadStrategy: aMaReadStrategy for: sessionId

	| session |
	session _ self sessionWithId: sessionId.
	aMaReadStrategy ifNotNil: [ session currentReadStrategy: aMaReadStrategy ].
	^session currentReadStrategy! !

!MaObjectRepository methodsFor: 'actions' stamp: 'cmm 12/16/2002 23:52'!
renewRefreshFlagFor: sessionId

	^(transactionLog entryFor: sessionId) renewRefreshFlag! !

!MaObjectRepository methodsFor: 'actions' stamp: 'cmm 1/25/2005 17:03'!
renewServerNotificationsIn: aMagmaServerResult for: sessionId

	self
		renewServerNotificationsIn: aMagmaServerResult
		for: sessionId
		from: (transactionLog entryFor: sessionId)! !

!MaObjectRepository methodsFor: 'actions' stamp: 'cmm 1/25/2005 17:02'!
renewServerNotificationsIn: aMagmaServerResult for: sessionId from: aMaTransactionLogEntry

	aMagmaServerResult
		serverNotifications: aMaTransactionLogEntry renewServerNotifications ;
		shouldRefresh: (self renewRefreshFlagFor: sessionId).
	aMagmaServerResult serverNotifications keysAndValuesDo:
		[ : eachKey : eachSet |
		aMagmaServerResult serverNotifications
			at: eachKey
			put: (eachSet collect: [ : eachOid | self byteArrayAt: eachOid using: MaReadStrategy deep ]) ]! !

!MaObjectRepository methodsFor: 'building' stamp: ''!
repositoryController: aMaRepositoryController

	repositoryController _ aMaRepositoryController.
	aMaRepositoryController repository == self ifFalse: [ aMaRepositoryController repository: self ]
! !

!MaObjectRepository methodsFor: 'private' stamp: 'cmm 12/31/2002 12:56'!
serverSerializer

	^repositoryController serverSerializer! !

!MaObjectRepository methodsFor: 'private' stamp: 'cmm 12/31/2002 12:55'!
session

	^repositoryController session! !

!MaObjectRepository methodsFor: 'private' stamp: 'cmm 4/28/2003 23:51'!
sessionWithId: anInteger

	^sessions 
		at: anInteger
		ifAbsent: [ MagmaUserError signal: 'session no longer valid.' ]! !

!MaObjectRepository methodsFor: 'private' stamp: 'cmm 5/2/2003 15:05'!
sessions

	^sessions! !

!MaObjectRepository methodsFor: 'actions' stamp: 'cmm 3/20/2005 20:52'!
submitAll: aMaCommitPackage for: sessionId beginAnother: aBoolean
	"Save all of the serialized objects in aMaCommitPackage to the repository.
	If any of the objects in aMaCommitPackage have been changed by other transactions
	signal a commit failure."

	| entry |
	entry _ self 
		commitEntryFor: sessionId 
		beginAnother: aBoolean.
	entry isCommitEntry ifFalse: [ MagmaUserError signal: 'No transaction to commit.' ].
	entry
		determineResultFor: aMaCommitPackage
		using: largeCollectionManagers.
	entry result hasConflicts
		ifTrue:
			[ MagmaCommitError new
				result: entry result;
				signal ]
		ifFalse:
			[ self
				writeAll: aMaCommitPackage
				recordPermanentOidsInResultFor: entry ].
	"Make sure system session doesn't exceed maximumNumberOfChallengers"
	self maMarked: 'performance'.  "the repositoryController session never enters a transaction, so it should be sensitive to the maximumNumberOfChallengers built-in mechanism for suggesting an abort.  Let's test that.."
	(sessionId ~= repositoryController sessionId
		and: [ transactionLog isFirstEntryFor: repositoryController sessionId ])
			ifTrue: [ repositoryController refreshView ].
	^entry result! !

!MaObjectRepository methodsFor: 'read strategies' stamp: 'cmm 9/1/2003 19:11'!
systemReadStrategy

	^systemReadStrategy! !

!MaObjectRepository methodsFor: 'magma-collections' stamp: 'cmm 3/18/2005 11:32'!
updateLargeCollectionsIn: aMaCommitPackage recordResultsIn: aMaCommitResult
	"The repositoryControllers sessions view of the repository must be completely refreshed, it must know about all new collections."

	| shouldEnsureOpen |
	shouldEnsureOpen _ false.
	aMaCommitPackage allLargeCollectionChangesDo:
		[ : each |
		each isNewCollection
			ifTrue:
				[ | permanentOid |
				permanentOid _ aMaCommitResult permanentOids at: each collectionOid ifAbsent: [ MagmaSoftwareError signal: 'expected to find permanent oid for new collection ', each collectionOid printString ].
				each 
					collectionOid: permanentOid ;
					createNewLargeCollectionUsing: self ].
		each hasAddedIndexes 
			ifTrue:
				[ each createSupplementaryIndexesUsing: self ].
		shouldEnsureOpen _ shouldEnsureOpen or: [ each isNewCollection or: [ each hasAddedIndexes ] ] ].
	shouldEnsureOpen ifTrue: [ repositoryController openLargeCollections ].
	aMaCommitPackage allLargeCollectionChangesDo:
		[ :each |
		(self collectionManagerFor: each collectionOid)
			process: each
			recordResultsIn: aMaCommitResult ]! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 4/21/2003 16:02'!
version

	^filer version! !

!MaObjectRepository methodsFor: 'private' stamp: 'cmm 3/20/2005 20:52'!
writeAll: aMaCommitPackage recordPermanentOidsInResultFor: aMaCommitLogEntry

	| newObjectOids |
	newObjectOids _
		self
			writeObjectsIn: aMaCommitPackage
			recordPermenantOidsIn: aMaCommitLogEntry result.
	aMaCommitLogEntry
		recordSuccessfullyCommittedIn: aMaCommitPackage
		using: newObjectOids.
	aMaCommitLogEntry hasAnyCommitted
		ifTrue:
			[ transactionLog challengeEntriesWith: aMaCommitLogEntry ].
	(aMaCommitPackage shouldServerRefresh or: [ aMaCommitLogEntry shouldServerRefresh ])
		ifTrue:
			[ repositoryController refreshView ].
	aMaCommitLogEntry didCommitSpecialOid
		ifTrue:
			[ self initializeClassDefinitionsSpecialOids ].
	self
		updateLargeCollectionsIn: aMaCommitPackage
		recordResultsIn: aMaCommitLogEntry result! !

!MaObjectRepository methodsFor: 'private' stamp: 'cmm 12/2/2004 13:29'!
writeObjectsIn: aMaCommitPackage recordPermenantOidsIn: aMaCommitResult

	| newObjectsOids |
	newObjectsOids _
		self
			linkNewObjectsIn: aMaCommitPackage
			recordPermenantOidsUsing: aMaCommitResult.
	aMaCommitPackage objectsDo:
		[ :eachBuffer |
		filer
			write: eachBuffer
			index: (newObjectsOids includes: eachBuffer oid) ].
	^ newObjectsOids! !

MaObject subclass: #MaTransactionLog
	instanceVariableNames: 'entries specialOids'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-private'!

!MaTransactionLog commentStamp: '<historical>' prior: 0!
This class resides on the server.
!

!MaTransactionLog methodsFor: 'actions' stamp: 'cmm 1/18/2005 22:43'!
abortTransactionEntryFor: sessionIdInteger repository: aMaObjectRepository

	^(self
		renewEntryFor: sessionIdInteger
		beginAnother: false
		for: aMaObjectRepository) result! !

!MaTransactionLog methodsFor: 'private' stamp: 'cmm 5/20/2004 01:04'!
challengeEntriesWith: aMaCommitLogEntry
	"We record all objects that were committed to all the entries in the receiver.  Some of the entries are in transaction, some are not.  The ones that are will use aMaCommitLogEntry to verify their commit has no conflicts with this entry.  The ones that are not will use the result of the receiver to know which objects need refreshed."

	| entriesToRemove |
	entriesToRemove _ nil.
	entries do:
		[ :each |
		(each isForSameSessionAs: aMaCommitLogEntry)
			ifFalse:
				[ each addChallengingEntry: aMaCommitLogEntry.
				each
					recordChangesToSpecialObjectsIn: aMaCommitLogEntry
					using: self.
				each hasExcessiveNumberOfChallengers
					ifTrue:
						[ self maMarked: 'dev'.
						"here is where we manage the health of the server.  Our current strategy is that your session 'times out' if you stay idle and get too far out of sync w/ the db.  Further work can be done here.  We should look into just marking this LogEntry 'bad' then, later, when they finally refresh, force it to refresh this sessions entire readSet..  something like that.."
						entriesToRemove ifNil: [ entriesToRemove _ OrderedCollection new ].
						entriesToRemove add: each.
						each needsToRefresh: true ] ] ].
	entriesToRemove ifNotNil:
		[ entriesToRemove 
			do:
				[ :each |
				self
					removeEntryForSessionId: each sessionId
					ifAbsent: [ "shouldn't happen" ] ] ]! !

!MaTransactionLog methodsFor: 'accessing' stamp: 'cmm 2/5/2003 20:46'!
entryFor: sessionId

	^entries
		detect: [ :each | each sessionId = sessionId ]
		ifNone: [ self signalLostSession: sessionId ]! !

!MaTransactionLog methodsFor: 'private' stamp: 'cmm 6/2/2004 12:25'!
initialize

	super initialize.
	"An OrderedCollection of MaCommitLogEntrys and MaTransactionLogEntrys.."
	entries _ OrderedCollection new.
	specialOids _ Dictionary new! !

!MaTransactionLog methodsFor: 'testing' stamp: 'cmm 12/31/2002 12:31'!
isFirstEntryFor: sessionId

	^entries notEmpty and: [ entries first connection sessionId = sessionId ]
! !

!MaTransactionLog methodsFor: 'building' stamp: 'cmm 4/28/2003 16:21'!
newSessionFor: userIdString numberOfChallengers: anInteger

	^(entries add: (
		MaTransactionLogEntry
			connection: (
				MagmaClientConnection
					userId: userIdString
					sessionId: self nextAvailableSessionId
					maximumNumberOfChallengers: anInteger)
			log: self)) connection! !

!MaTransactionLog methodsFor: 'building' stamp: 'cmm 1/28/2005 00:04'!
newTransactionFor: sessionIdInteger repository: aMaObjectRepository

	^ (self
		renewEntryFor: sessionIdInteger
		beginAnother: true
		for: aMaObjectRepository) result! !

!MaTransactionLog methodsFor: 'creation' stamp: ''!
nextAvailableSessionId

	^(entries
		inject: 0
		into: [ :highest :each | highest max: each sessionId ]) + 1
! !

!MaTransactionLog methodsFor: 'special oids' stamp: 'cmm 5/19/2004 22:24'!
onChangeOf: anInteger clientsShould: aSymbol

	specialOids
		at: anInteger
		put: aSymbol! !

!MaTransactionLog methodsFor: 'building' stamp: 'cmm 12/31/2002 12:44'!
removeEntryForSessionId: sessionId ifAbsent: aBlock

	^entries
		maDetect: [ :each | each sessionId = sessionId ]
		ifFound: [ :foundSession | entries remove: foundSession ]
		ifNone: aBlock! !

!MaTransactionLog methodsFor: 'actions' stamp: 'cmm 1/25/2005 17:04'!
renewEntryFor: sessionIdInteger beginAnother: aBoolean for: aMaObjectRepository
	"Gives the session identified by sessionIdInteger a new TranslactionLogEntry, where its challengers will be accumulated.  The entry prior to replacement is answered."

	| entry |
	entry _ self
		renewEntryFor: sessionIdInteger
		withNew: (aBoolean ifTrue: [ MaCommitLogEntry ] ifFalse: [ MaTransactionLogEntry ])
		ifAbsent: [ self signalLostSession: sessionIdInteger ].
	aMaObjectRepository
		renewServerNotificationsIn: entry result
		for: sessionIdInteger
		from: entry.
	^ entry! !

!MaTransactionLog methodsFor: 'private' stamp: 'cmm 4/21/2003 14:50'!
renewEntryFor: sessionIdInteger withNew: aClass ifAbsent: aBlock
	"aClass should be either MaTransactionLogEntry or MaCommitLogEntry"

	| entry |

	entry _
		(self
			removeEntryForSessionId: sessionIdInteger
			ifAbsent: [ ^aBlock value ]).
	entry recordAllChangedObjects.
	entries add: 
		(aClass
			connection: entry result connection
			log: self).
	^entry! !

!MaTransactionLog methodsFor: 'private' stamp: 'cmm 12/20/2002 12:43'!
signalLostSession: anInteger

	MagmaUserError signal:
		'Session ', anInteger printString, ' has timed out.  Did you stay idle for a long time?  Please reconnect.'! !

!MaTransactionLog methodsFor: 'special oids' stamp: 'cmm 5/19/2004 00:10'!
specialOidsDo: twoArgBlock

	specialOids keysAndValuesDo: twoArgBlock! !

MaObject subclass: #MaTransactionLogEntry
	instanceVariableNames: 'result challengingEntries needsToRefresh serverNotifications'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-private'!

MaTransactionLogEntry subclass: #MaCommitLogEntry
	instanceVariableNames: 'committed didCommitSpecialOid magmaArrayCommits'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-private'!

!MaCommitLogEntry methodsFor: 'private' stamp: ''!
committed

	^committed! !

!MaCommitLogEntry methodsFor: 'building' stamp: 'cmm 10/15/2004 16:57'!
determineResultFor: aMaCommitPackage using: lcManagersDictionary

	self
		recordLargeCollectionConflictsWith: aMaCommitPackage
			using: lcManagersDictionary;
		recordObjectConflictsIn: aMaCommitPackage.
	^ result! !

!MaCommitLogEntry methodsFor: 'testing' stamp: 'cmm 5/20/2004 21:39'!
didCommitSpecialOid

	^ didCommitSpecialOid! !

!MaCommitLogEntry methodsFor: 'private' stamp: 'cmm 10/15/2004 16:57'!
failResultWith: aMagmaClientConnection on: oidInteger

	result _
		result asFailedResult
			recordObjectConflictWith: aMagmaClientConnection
			on: oidInteger! !

!MaCommitLogEntry methodsFor: 'testing' stamp: 'cmm 3/22/2005 16:27'!
hasAnyCommitted

	^ committed notEmpty or: [ magmaArrayCommits notNil ]! !

!MaCommitLogEntry methodsFor: 'magmaArray support' stamp: 'cmm 3/20/2005 21:12'!
hasChangesToMagmaArray: oidInteger

	^ magmaArrayCommits notNil and:
	[ self magmaArrayCommits includesKey: oidInteger ]! !

!MaCommitLogEntry methodsFor: 'testing' stamp: 'cmm 10/15/2004 16:52'!
hasCommitted: oidInteger

	^ committed includesKey: oidInteger
! !

!MaCommitLogEntry methodsFor: 'initializing' stamp: 'cmm 5/20/2004 21:36'!
initialize

	super initialize.
	committed _ Dictionary new.
	didCommitSpecialOid _ false! !

!MaCommitLogEntry methodsFor: 'initializing' stamp: ''!
initializeResult

	result _ MaCommitResult new
! !

!MaCommitLogEntry methodsFor: 'testing' stamp: ''!
isCommitEntry

	^true
! !

!MaCommitLogEntry methodsFor: 'magmaArray support' stamp: 'cmm 3/20/2005 21:23'!
magmaArrayAdditionsAt: oidInteger

	^ self magmaArrayCommits
		at: oidInteger
		ifAbsentPut: [ Dictionary new ]! !

!MaCommitLogEntry methodsFor: 'magmaArray support' stamp: 'cmm 3/20/2005 21:08'!
magmaArrayCommits
	"A Dictionary, keyed by the oid of the MagmaArray that was changed; value is another Dictionary of the index->oids put into that array."

	^ magmaArrayCommits ifNil:
		[ magmaArrayCommits _ Dictionary new ]! !

!MaCommitLogEntry methodsFor: 'building' stamp: 'cmm 5/20/2004 21:38'!
noteDidCommitSpecialOid

	didCommitSpecialOid _ true! !

!MaCommitLogEntry methodsFor: 'private' stamp: 'cmm 3/16/2005 21:21'!
recordLargeCollectionConflictsWith: aMaCommitPackage using: lcManagersDictionary

	aMaCommitPackage allLargeCollectionChangesDo:
		[ : eachChanges |
		eachChanges isNewCollection ifFalse:
			[ | collectionManager |
			collectionManager _ lcManagersDictionary at: eachChanges collectionOid.
			collectionManager
				recordGeneralConflictsWith: eachChanges
				in: self.
			challengingEntries do:
				[ : eachEntry |
				collectionManager
					recordConflictsBetween: eachChanges
					and: eachEntry
					in: self ] ] ]! !

!MaCommitLogEntry methodsFor: 'private' stamp: 'cmm 10/15/2004 16:56'!
recordObjectConflictsIn: aMaCommitPackage
	"Go through everyone who has challenged me.  Determine if they have a same changed
	oid that I have.  Also fail if a new instance of a class is being added and shouldn't be."

	aMaCommitPackage persistentObjectsDo:
		[ :eachObjectBuffer |
		challengingEntries do:
			[ :eachEntry |
			(eachEntry hasCommitted: eachObjectBuffer oid)
				ifTrue:
					[ self
						failResultWith: eachEntry connection
						on: eachObjectBuffer oid ] ] ]! !

!MaCommitLogEntry methodsFor: 'private' stamp: 'cmm 10/8/2004 16:47'!
recordSuccessfulObjectCommit: aMaObjectBuffer

	committed
		at: aMaObjectBuffer oid
		put: aMaObjectBuffer! !

!MaCommitLogEntry methodsFor: 'building' stamp: 'cmm 3/20/2005 21:23'!
recordSuccessfullyCommittedIn: aMaCommitPackage using: newObjectOids
	"It is too late to utilize aMaCommitPackage>>newObjectsDo: because it has already been assigned its permanent oids and written out.  Now we must utilize the newObjectOoids Set to determine which buffers are for new objects."

	aMaCommitPackage objectsDo:
		[ : eachBuffer |
		(newObjectOids includes: eachBuffer oid)
			ifTrue:
				[ result recordNewObjectBuffer: eachBuffer copyWithSameBuffer ]
			ifFalse:
				[ self recordSuccessfulObjectCommit: eachBuffer copyWithSameBuffer ] ].
	result changedObjectBuffers: committed.
	aMaCommitPackage allLargeCollectionChangesDo:
		[ : eachChanges |
		eachChanges isForMagmaArray
			ifTrue:
				[ eachChanges additionsDo:
					[ : eachIndex : eachOid |
					(self magmaArrayAdditionsAt: eachChanges collectionOid)
						at: eachIndex
						put: eachOid ] ] ]! !

!MaCommitLogEntry methodsFor: 'testing' stamp: 'cmm 6/18/2004 15:52'!
shouldServerRefresh

	^ self didCommitSpecialOid! !

!MaCommitLogEntry methodsFor: 'accessing' stamp: 'cmm 10/15/2004 16:57'!
stopTime

	^ result stopTime
! !

!MaTransactionLogEntry class methodsFor: 'creation' stamp: 'cmm 5/2/2005 22:02'!
connection: aMagmaClientConnection log: aMaTransactionLog

	^self new 
		connection: aMagmaClientConnection ;
		yourself! !

!MaTransactionLogEntry methodsFor: 'building' stamp: 'cmm 10/15/2004 16:56'!
addChallengingEntry: aMaCommitLogEntry

	challengingEntries add: aMaCommitLogEntry! !

!MaTransactionLogEntry methodsFor: 'accessing' stamp: 'cmm 10/15/2004 16:58'!
connection

	^ result connection
! !

!MaTransactionLogEntry methodsFor: 'private' stamp: 'cmm 10/15/2004 16:58'!
connection: aMagmaClientConnection

	result connection: aMagmaClientConnection
! !

!MaTransactionLogEntry methodsFor: 'testing' stamp: 'cmm 10/15/2004 16:56'!
hasExcessiveNumberOfChallengers

	^ challengingEntries size > self connection maximumNumberOfChallengers
! !

!MaTransactionLogEntry methodsFor: 'private' stamp: 'cmm 3/20/2005 20:20'!
initialize
	"of MaTransactionLogEntries or MaCommitLogEntries that challenge me."

	super initialize.
	challengingEntries _ OrderedCollection new.
	self initializeResult.
	needsToRefresh _ false.
	self initializeServerNotifications! !

!MaTransactionLogEntry methodsFor: 'private' stamp: ''!
initializeResult

	result _ MaRefreshViewResult new
! !

!MaTransactionLogEntry methodsFor: 'actions' stamp: 'cmm 5/19/2004 00:20'!
initializeServerNotifications

	serverNotifications _ Dictionary new! !

!MaTransactionLogEntry methodsFor: 'testing' stamp: ''!
isCommitEntry

	^false
! !

!MaTransactionLogEntry methodsFor: 'testing' stamp: 'cmm 3/3/2002 19:00'!
isForSameSessionAs: aMaTransactionLogEntry

	^self connection == aMaTransactionLogEntry connection
! !

!MaTransactionLogEntry methodsFor: 'printing' stamp: 'cmm 6/20/2004 12:06'!
maPrintAbbreviatedOn: aStream

	aStream
		maPrint: '(' ;
		maPrint: self sessionId ;
		maPrint: ') ' ;
		maPrint: result userId! !

!MaTransactionLogEntry methodsFor: 'accessing' stamp: 'cmm 12/16/2002 23:28'!
needsToRefresh: aBoolean

	needsToRefresh _ aBoolean! !

!MaTransactionLogEntry methodsFor: 'accessing' stamp: 'cmm 10/15/2004 16:56'!
recordAllChangedObjects
	"Record all of the objects that the client will need to refresh."

	challengingEntries do: [ :each | result addToBeRefreshed: each committed ]
! !

!MaTransactionLogEntry methodsFor: 'actions' stamp: 'cmm 6/9/2004 16:58'!
recordChangesToSpecialObjectsIn: aMaCommitLogEntry using: aMaTransactionLog

	aMaTransactionLog specialOidsDo:
		[ : eachOid : eachKey |
		(aMaCommitLogEntry hasCommitted: eachOid)
			ifTrue:
				[ aMaCommitLogEntry noteDidCommitSpecialOid.
				(serverNotifications
					at: eachKey
					ifAbsentPut: [ Set new ]) add: eachOid ] ]! !

!MaTransactionLogEntry methodsFor: 'actions' stamp: 'cmm 12/25/2002 12:48'!
renewRefreshFlag

	| answer |
	answer _ needsToRefresh & self isCommitEntry not.
	self needsToRefresh: false.
	^answer! !

!MaTransactionLogEntry methodsFor: 'actions' stamp: 'cmm 5/19/2004 22:43'!
renewServerNotifications

	| answer |
	answer _ serverNotifications.
	serverNotifications notEmpty ifTrue: [ self initializeServerNotifications ].
	^ answer! !

!MaTransactionLogEntry methodsFor: 'private' stamp: ''!
result

	^result! !

!MaTransactionLogEntry methodsFor: 'accessing' stamp: 'cmm 10/15/2004 16:58'!
sessionId

	^ result sessionId
! !

!MaTransactionLogEntry methodsFor: 'accessing' stamp: 'cmm 10/15/2004 16:58'!
startTime

	^ result startTime
! !

MaObject subclass: #MagmaCompressionResult
	instanceVariableNames: 'unreferencedClasses unreferencedMagmaCollections'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-Utilities'!

!MagmaCompressionResult methodsFor: 'as yet unclassified' stamp: 'cmm 8/21/2002 21:44'!
unreferencedClasses

	^unreferencedClasses! !

!MagmaCompressionResult methodsFor: 'as yet unclassified' stamp: 'cmm 8/21/2002 21:45'!
unreferencedClasses: aCollection

	unreferencedClasses _ aCollection! !

!MagmaCompressionResult methodsFor: 'as yet unclassified' stamp: 'cmm 8/21/2002 21:45'!
unreferencedMagmaCollections

	^unreferencedMagmaCollections! !

!MagmaCompressionResult methodsFor: 'as yet unclassified' stamp: 'cmm 8/21/2002 21:45'!
unreferencedMagmaCollections: aCollection

	unreferencedMagmaCollections _ aCollection! !

MaObject subclass: #MagmaCompressor
	instanceVariableNames: 'inputController outputController referencedCollections referencedClasses classIdMap result sourceDirectory'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-Utilities'!

!MagmaCompressor commentStamp: 'cmm 8/29/2002 22:35' prior: 0!
I garbage collect a Magma repository and output a smaller file by eliminating buffers no longer referenced from the root buffer.

I always create a new object file, but if you use compressAndRename, I'll rename the source files to 'old...' so you can get with using the compressed repository right away.

If you don't want me to rename, just use #compressTo:.

Oh, and I'm not meant to work while the files are on-line.  For that, use:

	yourRepositoryController garbageCollect

(Note:  a potential strategy for on-line GC would be to set a boolean flag indicating to the filer not to reuse buffers made available through grown objects, but rather to always add to the end.  During this time, a background thread can start reading from the root and build a new copy of the 'memory' map (the MaIntervalCollection).  Once it has accounted for every byte in the real memory map, we know the real one can be replaced with this new one.  Do a quick write of the new memory map and you're done!!)!

!MagmaCompressor class methodsFor: 'creation' stamp: 'cmm 9/24/2004 16:49'!
inputController: aMagmaRepositoryController
	"NOT YET SUPPORTED.  The premise of this is that you can compress a running repository, but it is not complete.  Do not use!!"
	^self new inputController: aMagmaRepositoryController! !

!MagmaCompressor class methodsFor: 'creation' stamp: 'cmm 9/24/2004 16:49'!
source: pathString

	^self new sourceDirectory: (FileDirectory on: pathString)! !

!MagmaCompressor methodsFor: 'private' stamp: 'cmm 2/28/2005 15:38'!
buildClassIdMap
	"This should normally not be necessary.  But if Magma has been modified where an object is added to the repository definition, it's possible the initial class ids will not match.  This method builds a map of oldId->newId used when we write the new buffers."

	| inputClassMap outputClassMap |
	classIdMap _ Dictionary new.
	inputController requestCritical:
		[ inputClassMap _ inputController definition classDefinitions copy.
	outputClassMap _ outputController definition classDefinitions copy.
	inputClassMap keysAndValuesDo:
		[ : eachId : eachVersions |  | className |
		className _ (inputClassMap at: eachId) last className.
		"if output has consumed id for different class"
		((outputClassMap includesKey: eachId)
		and: [ (outputClassMap at: eachId) first className ~= eachVersions first className ])
			ifTrue:
				[ "try to find corresponding class-id in output, otherwise assign a new one."
				classIdMap
					at: eachId
					put:
						(outputClassMap
							maDetect: [ : eachDefinitions | eachDefinitions last className = className ]
							ifFound: [ : foundDefinitions | foundDefinitions last id ]
							ifNone: [ self nextAvailableClassId ]) ] ] ].
	classIdMap keysAndValuesDo:
		[ :eachOld :eachNew |
		MagmaNotification signal:
			'converting classId ', eachOld printString, 
			' to ', eachNew printString ]! !

!MagmaCompressor methodsFor: 'compress' stamp: 'cmm 2/28/2005 12:18'!
compressTo: outputDirectory

	| shouldCloseInput |
	shouldCloseInput _ inputController isNil.
	inputController ifNil: [ self openInput ].
	self
		createOutput: outputDirectory;
		openOutput: outputDirectory;
		ensureNoOidReuse ;
		buildClassIdMap;
		copyPrimitiveAttributes;
		writeGraphs ;
		copyMagmaCollections.
	outputController close.
	self removeUnreferencedCollectionsAndClasses: outputDirectory.
	shouldCloseInput ifTrue: [ inputController close ].
	^result
		unreferencedClasses: self unreferencedClasses;
		unreferencedMagmaCollections: self unreferencedMagmaCollections;
		yourself! !

!MagmaCompressor methodsFor: 'private' stamp: 'cmm 2/28/2005 16:30'!
copyMagmaCollections

	inputController repository magmaCollectionManagersDo:
		[ : each |
		(referencedCollections includes: each collectionOid)
			ifTrue:
				[ each filenames do:
					[ : eachMcFilename |
					inputController directory
						maCopyFileNamed: eachMcFilename
						toDirectory: outputController directory ] ] ]! !

!MagmaCompressor methodsFor: 'private' stamp: 'cmm 1/12/2005 21:13'!
copyPrimitiveAttributes

	| inputFiler outputFiler |
	inputFiler _ inputController repository filer.
	outputFiler _ outputController repository filer.
	inputController requestCritical:
		[ outputFiler
			signature: inputFiler signature ;
			"Even if this is a new version of Magma, the source buffers are still the old version."
			version: inputFiler version ;
			definitionOid: inputFiler definitionOid ;
			classDefinitionsOid: inputFiler classDefinitionsOid ;
			anchorOid: inputFiler anchorOid ]! !

!MagmaCompressor methodsFor: 'private' stamp: 'cmm 9/24/2004 12:24'!
createOutput: aString

	MagmaRepositoryController create: aString! !

!MagmaCompressor methodsFor: 'private' stamp: 'cmm 2/17/2005 13:48'!
ensureNoOidReuse
	"We can never allow an oid to be reused.  Once an object is created and persisted, its oid must be forever consumed.  Even if that object is later dereferenced, its oid will never be reused.  This is necessary to properly support MagmaForwardingProxys."

	| highestOid |
	"determine the next oid to be used in our inputController."
	highestOid _ (inputController repository filer getNextOid) - 1.
	"Force our output controller to allocate up to the current oid."
	outputController repository filer
		index: highestOid
		at: 0! !

!MagmaCompressor methodsFor: 'initialize-release' stamp: 'cmm 12/24/2002 15:51'!
initialize
	super initialize.
	referencedClasses := Set new.
	referencedCollections := Set new.
	result := MagmaCompressionResult new! !

!MagmaCompressor methodsFor: 'initialize-release' stamp: 'cmm 8/17/2002 15:25'!
inputController: aMagmaRepositoryController

	inputController _ aMagmaRepositoryController! !

!MagmaCompressor methodsFor: 'private' stamp: 'cmm 5/2/2005 19:16'!
markClassReferenced: aMaObjectBuffer 
	| classIdManager |
	classIdManager := inputController session serializer classIdManager.
	(classIdManager 
		definitionForClassId: aMaObjectBuffer classId
		version: aMaObjectBuffer classVersion) 
		withAllSuperclassDefinitionsDo: 
			[ : eachClassDefinition | 
			referencedClasses add: (self newClassIdFor: eachClassDefinition id) ].
	"If a class-reference, be sure to get the referenced class itself."
	(aMaObjectBuffer classId = (classIdManager idForClass: MaClassReference))
		ifTrue:
			[ | indexOfId classReferenceDefinition |
			classReferenceDefinition _
				classIdManager
					definitionForClassId: aMaObjectBuffer classId
					version: aMaObjectBuffer classVersion.
			"Get the index of the 'id' variable."
			indexOfId _ classReferenceDefinition allInstVarNames indexOf: 'id'.
			indexOfId = 0 ifTrue: [ MagmaSoftwareError signal: 'Expected ''id'' variable.' ].
			(classIdManager 
				definitionForClassId:
					(inputController session
						objectWithOid: (aMaObjectBuffer maInstVarAt: indexOfId)
						ifAbsent: [ MagmaSoftwareError signal: 'Expected an to find an id' ])
				version: 1) 

				withAllSuperclassDefinitionsDo: 
					[ : eachClassDefinition | 
					referencedClasses add: (self newClassIdFor: eachClassDefinition id) ] ]! !

!MagmaCompressor methodsFor: 'private' stamp: 'cmm 8/22/2002 21:09'!
markCollectionReferenced: oidInteger

	referencedCollections add: oidInteger! !

!MagmaCompressor methodsFor: 'private' stamp: 'cmm 7/2/2004 00:07'!
newClassIdFor: sourceClassId

	^ classIdMap at: sourceClassId ifAbsent: [ sourceClassId ]! !

!MagmaCompressor methodsFor: 'private' stamp: 'cmm 7/1/2004 17:22'!
nextAvailableClassId
	"Answer the lowest commonly-available classId between the input and outputController."

	^ inputController requestCritical:
		[ | inputClassMap outputClassMap |
		inputClassMap _ inputController definition classDefinitions.
		outputClassMap _ outputController definition classDefinitions.
		(inputClassMap keys max max: outputClassMap keys max) + 1 ]! !

!MagmaCompressor methodsFor: 'private' stamp: 'cmm 9/24/2004 16:54'!
openInput

	inputController _ MagmaRepositoryController open: sourceDirectory pathName! !

!MagmaCompressor methodsFor: 'private' stamp: 'cmm 8/14/2002 23:41'!
openOutput: aString

	outputController _ MagmaRepositoryController open: aString! !

!MagmaCompressor methodsFor: 'private' stamp: 'cmm 2/23/2005 15:04'!
removeUnreferencedCollectionsAndClasses: outputDirectory

	| controller |
	controller _ outputController privateOpen: (FileDirectory on: outputDirectory).
	controller session commit:
		[ controller definition classDefinitions copy
			keysDo: 
				[ :eachId |
				(referencedClasses includes: eachId) ifFalse:
					[ controller definition removeClassWithId: eachId ] ].
		controller definition allLargeCollections copy do:
			[ :eachCollection |
			(referencedCollections includes: (controller session oidFor: eachCollection)) ifFalse:
				[ controller definition removeCollection: eachCollection ] ] ].
	controller close! !

!MagmaCompressor methodsFor: 'initialize-release' stamp: 'cmm 9/24/2004 16:50'!
sourceDirectory: aFileDirectory

	sourceDirectory _ aFileDirectory! !

!MagmaCompressor methodsFor: 'accessing' stamp: 'cmm 7/1/2004 23:35'!
unreferencedClasses

	^ outputController definition classDefinitions reject:
		[ : eachDefinitions |
		referencedClasses includes: eachDefinitions first id ]! !

!MagmaCompressor methodsFor: 'accessing' stamp: 'cmm 3/20/2005 19:24'!
unreferencedMagmaCollections

	| unreferenced |
	unreferenced _ inputController definition allLargeCollections reject: [ :each | referencedCollections includes: (inputController session oidFor: each) ].
	^ unreferenced collect: 
		[ : eachCollection | 
		inputController repository 
			collectionFilenameFor: eachCollection changes collectionOid
			extension: eachCollection changes suggestedFileExtension ]! !

!MagmaCompressor methodsFor: 'private' stamp: 'cmm 5/2/2005 14:07'!
writeGraphs

	| inputTraverser |
	inputTraverser _ MagmaFileTraverser repositoryController: inputController.
	inputTraverser anchorGraphDo:
		[ : eachBuffer |
		self markClassReferenced: eachBuffer.
		inputTraverser inDefinition ifFalse:
			[ (inputTraverser isOidForLargeCollection: eachBuffer oid)
				ifTrue:
					[ self markCollectionReferenced: eachBuffer oid ] ].
		eachBuffer classId: (self newClassIdFor: eachBuffer classId).
		"If not in a buffer embedded in a MaStorageObjectBuffer..."
		(MaOidCalculator isOidForUserObject: eachBuffer oid)
			ifTrue:
				[ outputController requestCritical:
					[ outputController repository filer
						write: eachBuffer
						index: true ] ] ]! !

MaObject subclass: #MagmaFileTraverser
	instanceVariableNames: 'repositoryController oidsVisited graphBuffer directory inDefinition'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-Utilities'!

!MagmaFileTraverser commentStamp: 'cmm 8/5/2002 21:45' prior: 0!
I walk the object buffers of an open Magma file and execute a Block for each buffer.

I'm useful for the Magma garbage collection function, the compression function and for general purpose information gathering about what's in the repository.!

!MagmaFileTraverser class methodsFor: 'creation' stamp: 'cmm 9/24/2004 15:56'!
for: pathString

	^ self new 
		directory: (FileDirectory on: pathString) ;
		yourself! !

!MagmaFileTraverser class methodsFor: 'creation' stamp: 'cmm 8/15/2002 22:52'!
repositoryController: aMagmaRepositoryController

	^self new repositoryController: aMagmaRepositoryController! !

!MagmaFileTraverser methodsFor: 'actions' stamp: 'cmm 5/2/2005 10:37'!
anchorGraphDo: aBlock

	self
		graphFrom: (repositoryController requestCritical: [ self repository filer anchorOid ])
		do: aBlock! !

!MagmaFileTraverser methodsFor: 'private' stamp: 'cmm 8/21/2002 00:24'!
close

	repositoryController ifNotNil:
		[ repositoryController requestCritical: [ repositoryController close ] ]! !

!MagmaFileTraverser methodsFor: 'private' stamp: 'cmm 9/24/2004 16:15'!
deleteOidsVisited

	oidsVisited ifNotNil: [ oidsVisited close ].
	directory deleteFileNamed: self tempOidsVisitedFilename! !

!MagmaFileTraverser methodsFor: 'actions' stamp: 'cmm 5/2/2005 10:37'!
graphFrom: oid do: aBlock
	"Value aBlock against every buffer in the receivers file."

	| readStrategy shouldClose |
	shouldClose _ false.
	repositoryController ifNil: [ shouldClose _ true. self open ].
	readStrategy _ MaBasicReadStrategy minimumDepth: 0.
	MagmaNotification signal: 'Beginning traversal of ', oid printString, ' of ', directory pathName.
	graphBuffer _ repositoryController session serializer graphBuffer copy resetBufferSize; yourself.  "sorry Demeter"
	[ self initializeOidsVisited.
	self
		graphFrom: oid
		do: aBlock
		using: readStrategy ]

		ensure:
			[ shouldClose ifTrue: [ self close ].
			self deleteOidsVisited ].
	MagmaNotification signal: 'Traversal of ', oid printString, ' of ', directory pathName, ' complete.'! !

!MagmaFileTraverser methodsFor: 'private' stamp: 'cmm 5/2/2005 10:36'!
graphFrom: oidInteger do: oneArgBlock using: aMaReadStrategy

	| buffer indexPosition |
	(MaOidCalculator isOidForUserObject: oidInteger) ifFalse: [ ^self ].
	indexPosition _ MaOidCalculator indexPositionForOid: oidInteger.
	(oidsVisited at: indexPosition ifAbsent: [ 0 ]) = 0 ifFalse: [ ^self ].
	"Don't enumerate the MagmaCollections referenced from the definition, just the ones referenced from the root, that way we can properly calculate which MagmaCollections are still referenced and which are not."
	(inDefinition and: [ self isOidForLargeCollection: oidInteger ])
		ifFalse:
			[ oidsVisited
				at: indexPosition
				put: oidInteger ].
	repositoryController requestCritical:
		[ buffer _ self repository filer
			appendObject: oidInteger
			into: graphBuffer ].
	oneArgBlock value: buffer.
	(buffer oid = repositoryController repository filer definitionOid) 
		ifTrue: [ inDefinition _ true ].
	buffer
		graphDo: oneArgBlock
		using: aMaReadStrategy
		for: self
		inDefinition: inDefinition.
	buffer oid = repositoryController repository filer definitionOid
		ifTrue: [ inDefinition _ false ].
	graphBuffer decreaseBufferSizeBy: buffer physicalSize! !

!MagmaFileTraverser methodsFor: 'private' stamp: 'cmm 5/1/2005 18:22'!
graphFromLargeCollection: oidInteger do: oneArgBlock using: aMaReadStrategy
	"aMaByteObjectBuffer is for a storageObject.  Therefore enumerate its buffers because its objects are part of the repository too."

	| fileIndex |
	fileIndex _ (repositoryController 
		requestCritical: 
			[ repositoryController largeCollectionManagerFor: oidInteger ]) memberIndex.
	MagmaNotification signal: 'beginning enumeration of ', fileIndex size printString, ' objects in MagmaCollection ', oidInteger printString.
	fileIndex
		do: oneArgBlock
		for: self
		using: aMaReadStrategy.
	MagmaNotification signal: 'MagmaCollection enumeration complete:  ', oidInteger printString! !

!MagmaFileTraverser methodsFor: 'private' stamp: 'cmm 5/2/2005 12:04'!
graphFromPointers: aMaObjectBuffer do: oneArgBlock using: aMaReadStrategy
	"aMaObjectBuffer isPointers must be true to get here."

	aMaObjectBuffer instVarsDoWithIndex:
		[ :eachOid :index |
		self
			graphFrom: eachOid
			do: oneArgBlock
			using: aMaReadStrategy ]! !

!MagmaFileTraverser methodsFor: 'testing' stamp: 'cmm 2/28/2005 13:40'!
inDefinition
	"Answer whether I am currently traversing inside my input-controller's repository-definition."

	^ inDefinition! !

!MagmaFileTraverser methodsFor: 'private' stamp: 'cmm 2/28/2005 13:40'!
initialize
	super initialize.
	inDefinition _ false! !

!MagmaFileTraverser methodsFor: 'private' stamp: 'cmm 4/28/2005 20:40'!
initializeOidsVisited

	oidsVisited ifNotNil: [ self deleteOidsVisited ].
	MaLargeArrayOfNumbers
		create: self tempOidsVisitedFilename
		bitSize: MaObjectFiler maxDbSize.
	oidsVisited _ MaLargeArrayOfNumbers open: self tempOidsVisitedFilename! !

!MagmaFileTraverser methodsFor: 'private' stamp: 'cmm 3/22/2005 17:55'!
isOidForLargeCollection: oidInteger

	^ self repository largeCollectionManagers includesKey: oidInteger! !

!MagmaFileTraverser methodsFor: 'private' stamp: 'cmm 9/24/2004 15:52'!
open
	"Use if you created with just a path, or if you know the MagmaRepositoryController needs opened.."

	repositoryController _ MagmaRepositoryController open: directory pathName! !

!MagmaFileTraverser methodsFor: 'private' stamp: 'cmm 8/13/2002 23:29'!
repository

	^repositoryController repository! !

!MagmaFileTraverser methodsFor: 'accessing' stamp: 'cmm 1/19/2005 21:36'!
repositoryController

	^ repositoryController! !

!MagmaFileTraverser methodsFor: 'private' stamp: 'cmm 9/24/2004 15:54'!
repositoryController: aMagmaRepositoryController

	repositoryController _ aMagmaRepositoryController.
	directory _ aMagmaRepositoryController directory! !

!MagmaFileTraverser methodsFor: 'private' stamp: 'cmm 9/26/2004 21:30'!
tempOidsVisitedFilename

	^ repositoryController directory fullNameFor: 'oidsVisited.tmp'! !

MaObject subclass: #MagmaRepositoryController
	instanceVariableNames: 'repository session serverSerializer requestInterruptGuard isReadOnly directory'
	classVariableNames: 'OpenControllers'
	poolDictionaries: ''
	category: 'Magma server-API'!

!MagmaRepositoryController class methodsFor: 'utilities' stamp: 'cmm 9/24/2004 17:00'!
compress: sourcePath to: targetPath

	(MagmaCompressor source: sourcePath) compressTo: targetPath! !

!MagmaRepositoryController class methodsFor: 'accessing' stamp: 'cmm 2/20/2005 16:29'!
controllerOpenedOn: pathName

	^ OpenControllers
		maDetect: [ : each | each directory pathName = pathName ]
		ifFound: [ : foundController | foundController ensureOpen ]
		ifNone: [ nil ]! !

!MagmaRepositoryController class methodsFor: 'utilities' stamp: 'cmm 9/26/2004 17:42'!
create: fullyQualifiedPathString
	"Creates a repository with an initial root of nil.  I #assureExistence of the fullyQualifiedPathString."

	| fd |
	self validateCanCreateIn: (fd _ FileDirectory on: fullyQualifiedPathString).
	self new createRepositoryAt: fd! !

!MagmaRepositoryController class methodsFor: 'utilities' stamp: 'cmm 1/16/2005 21:54'!
create: fullyQualifiedPathString root: anObject

	| session |
	self create: fullyQualifiedPathString.
	session _ MagmaSession openLocal: fullyQualifiedPathString.
	session
		connect: (MagmaUser id: 'system create:root:') ;
		commit: [ session root: anObject ];
		disconnect;
		closeRepository! !

!MagmaRepositoryController class methodsFor: 'utilities' stamp: 'cmm 3/11/2005 14:57'!
delete: fullyQualifiedPathString
	"Delete the just the files in fullyQualifiedPathString that are part of the Magma repository that resides there."

	| fd |
	((fd _ FileDirectory on: fullyQualifiedPathString) exists
	and: [ fd fileExists: MaObjectFiler objectsFilename ])
		ifFalse:
			[ (MagmaNotification new messageText: fullyQualifiedPathString, ' does not exist, nothing to delete.') signal.
			^ self ].
	(self filesFor: fullyQualifiedPathString) do:
		[ :each |
		(FileDirectory on: fullyQualifiedPathString)
			deleteFileNamed: each asString
			ifAbsent:
				[ MaNotification signal: 'Couldn''t find ', each asString ] ]! !

!MagmaRepositoryController class methodsFor: 'private' stamp: 'cmm 8/22/2002 20:52'!
filesFor: filename

	| controller |
	controller _ self open: filename.
	^[ controller filenames ] ensure: [ controller close ]! !

!MagmaRepositoryController class methodsFor: 'initializing' stamp: 'cmm 11/2/2004 14:38'!
initialize

	OpenControllers ifNotNil: [ OpenControllers do: [ : each | each close ] ].
	OpenControllers _ Set new.
	Smalltalk
		addToStartUpList: self;
		addToShutDownList: self! !

!MagmaRepositoryController class methodsFor: 'private' stamp: 'cmm 9/26/2004 19:21'!
open: fullyQualifiedPathString

	^ self new open: (FileDirectory on: fullyQualifiedPathString)! !

!MagmaRepositoryController class methodsFor: 'accessing' stamp: 'cmm 10/26/2004 12:48'!
openedOn: fullyQualifiedPathString

	| existingController |
	^ (existingController _ self controllerOpenedOn: fullyQualifiedPathString)
		ifNil: [ self open: fullyQualifiedPathString ]
		ifNotNil: [ existingController ]! !

!MagmaRepositoryController class methodsFor: 'private' stamp: 'cmm 5/1/2005 23:21'!
protocol
	"This is the protocol needed to talk to a magma server."
	^ { 
	"core"
		LargePositiveInteger. MaSerializedGraphBuffer. 
	"part of MagmaCollectionChanges"
		IdentityDictionary.  Association.  MaChangedKeySpecification.
	"extra objects for some requests."
		MaCommitPackage.  MaBasicReadStrategy.  MaReadStrategy.
	"various responses"
		MagmaId. UUID. MaReadResult. MaFailedCommitResult. MagmaClientConnection.  MagmaLargeCollectionSegment. MaRefreshViewResult. MaCommitResult. MagmaCommitConflict }, 
	"exceptions"
		MaError allSubclasses asArray,
	"various requests"
		MagmaRepositoryRequest allSubclasses asArray! !

!MagmaRepositoryController class methodsFor: 'utilities' stamp: 'cmm 9/24/2004 13:01'!
repositoryExistsIn: aFileDirectory

	^ (aFileDirectory directoryEntryFor: MaObjectFiler objectsFilename) notNil! !

!MagmaRepositoryController class methodsFor: 'initializing' stamp: 'cmm 12/27/2004 23:16'!
shutDown: aboutToQuit

	super shutDown: aboutToQuit.
	MagmaSession shutDown: aboutToQuit.
	OpenControllers
		do:
			[ :each | 
			[ MagmaNotification signal: 'closing Magma repository ', each pathName asString, '.  Will reopen on next startup.' ]
				on: MagmaNotification
				do: [ : noti | Transcript cr; show: noti messageText. noti pass ].
			each primClose ]! !

!MagmaRepositoryController class methodsFor: 'initializing' stamp: 'cmm 3/11/2005 14:58'!
startUp: amResuming

	super startUp: amResuming.
	OpenControllers copy do: 
		[ :each | 
		[ [ each ensureOpen ]
			on: MagmaNotification
			do: [ : noti | Transcript cr; show: noti messageText. noti pass ] ]
				on: MagmaEnvironmentError
				do: 
					[ : error | "not present anymore, remove from OpenControllers and resignal."
					each close.
					error pass ] ].
	"Now that we're open, ensure any server-sockets that may need to restart can be."
	MaServerSocket startUp: amResuming! !

!MagmaRepositoryController class methodsFor: 'private' stamp: 'cmm 5/1/2003 00:24'!
systemSessionUserId

	^'__system'! !

!MagmaRepositoryController class methodsFor: 'validation' stamp: 'cmm 9/24/2004 12:32'!
validateCanCreateIn: aFileDirectory
	"Magma databases do not share the same directory.  Not only would it be hard to identify which files go with which repository, some of the filenames are based on oids, which would not always be unique across repositories."

	(aFileDirectory exists and: [ self repositoryExistsIn: aFileDirectory ])
		ifTrue:
			[ MagmaUserError signal: aFileDirectory pathName, ' already has a Magma repository in it.  A directory may contain only one Magma repository.' ]
! !

!MagmaRepositoryController methodsFor: 'client requests' stamp: ''!
abortTransactionFor: sessionId

	^self repository abortTransactionFor: sessionId
! !

!MagmaRepositoryController methodsFor: 'statistics' stamp: 'cmm 4/21/2003 16:28'!
averageObjectSize

	^repository objectFileSize // self objectCount! !

!MagmaRepositoryController methodsFor: 'security' stamp: 'cmm 4/30/2003 17:21'!
beReadOnly

	self requestCritical: [ isReadOnly _ true ]! !

!MagmaRepositoryController methodsFor: 'security' stamp: 'cmm 4/30/2003 17:22'!
beReadWrite

	self requestCritical: [ isReadOnly _ false ]! !

!MagmaRepositoryController methodsFor: 'accessing' stamp: 'cmm 10/24/2004 23:14'!
close

	OpenControllers remove: self ifAbsent: [ "do nothing" ].
	self primClose! !

!MagmaRepositoryController methodsFor: 'client requests' stamp: 'cmm 10/26/2004 21:54'!
codeBaseByteArray

	^Array
		with: repository classDefinitionsByteArray
		with: (serverSerializer serializeGraph: self definition codeBase) trimmedByteArray! !

!MagmaRepositoryController methodsFor: 'initializing' stamp: 'cmm 1/16/2005 21:54'!
connect

	session connect: (MagmaUser id: self class systemSessionUserId).
	repository initializeSpecialOidsList! !

!MagmaRepositoryController methodsFor: 'unsupported' stamp: 'cmm 6/23/2002 10:48'!
connections
	"Used for monitoring server status via some console gui."

	self maMarked: 'dev'.! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 12/29/2004 00:22'!
createRepositoryAt: aFileDirectory

	| anchor |
	directory _ aFileDirectory.
	self
		repository: (
			MaObjectRepository
				create: aFileDirectory
				controller: self);
		connect;
		forceWrite:
			(anchor _ MaRootAnchor new
				definition: self definition ;
				yourself).
	repository filer
		anchorOid: (self session oidFor: anchor);
		classDefinitionsOid: (self session oidFor: self definition classDefinitions);
		definitionOid: (self session oidFor: self definition).
	self close! !

!MagmaRepositoryController methodsFor: 'accessing' stamp: 'cmm 5/2/2004 20:40'!
definition
	"Answers a MagmaRepositoryDefinition"

	^ session definition! !

!MagmaRepositoryController methodsFor: 'accessing' stamp: 'cmm 9/24/2004 15:55'!
directory
	"Answer the directory in which my files reside."
	"Note, this is the same FileDirectory that MaObjectFiler has.  It was necessary to duplicate this here to support the auto-reopen feature upon image restart."
	^ directory! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 12/26/2004 22:38'!
disconnect: sessionId

	repository removeSessionFor: sessionId.
	^nil "answer nil to minimize the serialized response."! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 2/20/2005 16:31'!
ensureOpen

	self isOpen ifFalse: [ self open ]! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 9/1/2003 19:10'!
ensureReadStrategy: aMaReadStrategy forSession: sessionId

	| strategy |
	^(strategy _
		repository
			renewCurrentReadStrategy: aMaReadStrategy
			for: sessionId)

				ifNil: [ repository systemReadStrategy ] 
				ifNotNil: [ strategy ]! !

!MagmaRepositoryController methodsFor: 'accessing' stamp: 'cmm 9/26/2004 21:04'!
filenames

	^ { MaObjectFiler objectsFilename.  MaObjectFiler objectPointersFilename },
	(repository largeCollectionManagers 
		inject: OrderedCollection new
		into:
			[ : coll : each |
			coll
				addAll: each filenames ;
				yourself ])! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 12/29/2004 00:09'!
forceWrite: anObject
	"This is somewhat of a hacking method to enable us to write objects needed
	for the repository to operate (i.e., commit transactions) during initial creation."

	| commitPackage |
	session begin.  "Begin a tran to allow the submitAll:for:.."
	commitPackage _ MaCommitPackage new.
	session serializer
		serializeGraph: anObject
		do: [ :eachObject | commitPackage addObject: eachObject ].
	commitPackage copy serializeObjectsUsing: session.
	commitPackage serializeObjectsUsing: session.
	session
		refreshViewUsing: (
			self
				submitAll: commitPackage
				for: session id
				beginAnother: false) ;
		decrementTransactionLevel! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 4/3/2003 23:00'!
getDefinition

	^session materializeObject: repository definitionByteArray! !

!MagmaRepositoryController methodsFor: 'initializing' stamp: 'cmm 1/16/2005 18:45'!
initialize

	super initialize.
	self initializeSession.
	session initializeDefinition.
	session definition addMinimumClassesForOdbmsUsing: self serverSerializer.
	requestInterruptGuard _ Semaphore forMutualExclusion.
	isReadOnly _ false! !

!MagmaRepositoryController methodsFor: 'initializing' stamp: 'cmm 3/27/2002 22:00'!
initializeServerSerializer
	"Why is this separate, 'serverSerializer' needed?  Because the session serializer requires a different traversalStrategy.  The serverSerializer needs to be able to fully serialize, to the ends of the graphs, the 'response' objects of various requests."

	serverSerializer _ self session serializer copyWithNewBuffer initializeTraversalStrategy
! !

!MagmaRepositoryController methodsFor: 'initializing' stamp: 'cmm 1/16/2005 18:45'!
initializeSession

	session _ MagmaSession repositoryController: self.
	self initializeServerSerializer! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 9/1/2003 19:40'!
initializeSystemReadStrategy

	repository initializeSystemReadStrategyUsing: self serverSerializer classIdManager! !

!MagmaRepositoryController methodsFor: 'accessing' stamp: 'cmm 4/30/2003 22:07'!
isOpen

	^repository isOpen! !

!MagmaRepositoryController methodsFor: 'security' stamp: 'cmm 12/17/2002 14:45'!
isReadOnly

	^isReadOnly! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 4/18/2005 15:00'!
largeCollectionManagerFor: collectionOidInteger

	^ repository largeCollectionManagers
		at: collectionOidInteger
		ifAbsent:
			[ MagmaSoftwareError signal: 'No manager exists for ', collectionOidInteger printString ]! !

!MagmaRepositoryController methodsFor: 'accessing' stamp: 'cmm 1/14/2005 16:05'!
magmaId

	^ self definition magmaId! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 6/14/2004 18:13'!
materializeClassDefinitions
	"This is needed because, when opening the repository, the default classes do not include MagmaRepositoryDefinition, which is delivered as part of any session-connection.  When the system-session attempts to connect, it tries to read the repository-definition but can't because it can't get a proper read-strategy for it."

	session loadClassDefinitionsFrom: repository classDefinitionsByteArray! !

!MagmaRepositoryController methodsFor: 'unsupported' stamp: 'cmm 5/2/2004 20:40'!
maximumNumberOfChallengers: anInteger 
	self session 
		commit: [ self definition maximumNumberOfChallengers: anInteger ]! !

!MagmaRepositoryController methodsFor: 'client requests' stamp: 'cmm 1/16/2005 21:47'!
newSessionFor: userIdString

	^ repository
		newSessionFor: userIdString
		numberOfChallengers: MagmaRepositoryDefinition defaultMaximumNumberOfChallengers! !

!MagmaRepositoryController methodsFor: 'client requests' stamp: 'cmm 12/26/2004 21:12'!
newTransactionFor: sessionId

	| refreshViewResult |
	isReadOnly ifTrue:
		[ MagmaUserError signal: 'This repository is set to read-only, commits are prohibited.' ].
	refreshViewResult _ repository newTransactionFor: sessionId.
	^ refreshViewResult! !

!MagmaRepositoryController methodsFor: 'statistics' stamp: 'cmm 5/2/2003 14:56'!
numberOfConnectedClients

	^repository sessions size - 1! !

!MagmaRepositoryController methodsFor: 'client requests' stamp: 'cmm 5/1/2002 22:55'!
numberOfEntriesFrom: lowKeyInteger to: highKeyInteger inLargeCollection: collectionOid on: attributeSymbol

	| index |

	index _
		(self largeCollectionManagerFor: collectionOid)
			indexForAttribute: attributeSymbol
			ifNone:
				[ MagmaSoftwareError signal:
					attributeSymbol , ' is not an index of large-collection with oid ' ,
						collectionOid printString , '.' ].
	^index
		numberOfEntriesFrom: lowKeyInteger
		to: highKeyInteger
! !

!MagmaRepositoryController methodsFor: 'statistics' stamp: 'cmm 4/21/2003 15:55'!
objectCount
	"Answers how many objects have been committed to this repository.  Does not account for objects that have since been garbage colellected."

	^repository objectCount! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 9/24/2004 13:38'!
open
	"This is called when restarting controllers during image restart."

	self
		privateOpen: directory ;
		openLargeCollections! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 9/24/2004 13:32'!
open: aFileDirectory

	OpenControllers add: self. "<- we only do this for intentional opens, not image restarts"
	directory _ aFileDirectory.
	self open! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 3/16/2005 12:46'!
openLargeCollection: aMagmaLargeCollection

	repository openLargeCollectionIfNecessary: aMagmaLargeCollection
! !

!MagmaRepositoryController methodsFor: 'initializing' stamp: 'cmm 5/2/2004 20:40'!
openLargeCollections
	"Open all indexes for all large collections."

	self definition allLargeCollections 
		do: [ :each | self openLargeCollection: each ]! !

!MagmaRepositoryController methodsFor: 'client requests' stamp: 'cmm 9/24/2004 15:13'!
pathName

	^ directory pathName! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 3/24/2005 09:51'!
primClose

"Here, it should shutdown gracefully, but any queued requests will not be processed.  Once VI4 is available, we can uncomment the version below for more grace."
	self requestCritical:
		[ session ifNotNil: [ session disconnect ].
		repository close ].

self maMarked: 'dev'.
"once we get block-closures, we should be able to safely close the repository using the code below.  The code below utilizes the existing client server framework's request queueing to ensure all pending requests are processed before we close, ensuring everyones work gets saved.  Unfortunately, it requires a double-evaluation of the processBlock.  Until then, you'll have to exercise your best judgement."


	"go through the session to close the repository so we ensure all pending requests are processed."
"
	[ session closeRepository ] fork.
	[ self isOpen ] whileTrue: [ (Delay forSeconds: 1) wait ]
"! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 2/20/2005 16:27'!
privateOpen: aFileDirectory

	MagmaNotification signal: 'opening Magma repository ', self pathName.
	self
		initializeSession;
		repository: (
			MaObjectRepository
				open: aFileDirectory
				controller: self) ;
		initializeSystemReadStrategy ;
		materializeClassDefinitions ;
		connect ;
		refreshView "so more classes will be pulled from unusedDepths of the SystemReadStrategy"! !

!MagmaRepositoryController methodsFor: 'client/server required methods' stamp: 'cmm 4/10/2005 18:00'!
processRequest: aMaRepositoryRequest 
	aMaRepositoryRequest repositoryController: self.
	^
	[aMaRepositoryRequest sessionId = self sessionId 
		ifTrue: [aMaRepositoryRequest process]
		ifFalse: [self requestCritical: [aMaRepositoryRequest process]]] 
			on: MaError
			do: 
				[:anError | 
				(MagmaPreferences debug and: [anError maOriginalClass ~= MagmaCommitError]) 
					ifTrue: [self halt: anError messageText].
				anError]! !

!MagmaRepositoryController methodsFor: 'client/server required methods' stamp: 'cmm 5/1/2005 23:21'!
protocol

	| answer |

	answer _ self class protocol asOrderedCollection.
	answer 
		addAll: MaObjectBuffer allSubclasses ;
		addAll: MaIndexDefinition allSubclasses ;
		addAll: MagmaLargeCollectionChanges allSubclasses.
	(Smalltalk hasClassNamed: #MagmaTestCase)
		ifTrue: [ answer addAll: (Smalltalk at: #MagmaTestCase) additionalProtocolForRepositoryController ].
	^answer
! !

!MagmaRepositoryController methodsFor: 'client requests' stamp: 'cmm 1/25/2005 15:55'!
read: oidInteger for: sessionId using: aMaReadStrategy

	| result |
	result _ MaReadResult byteArray:
		(repository
			byteArrayAt: oidInteger
			using:
				(self 
					ensureReadStrategy: aMaReadStrategy
					forSession: sessionId)).
	repository
		renewServerNotificationsIn: result
		for: sessionId.
	^result! !

!MagmaRepositoryController methodsFor: 'client requests' stamp: 'cmm 2/9/2005 10:42'!
readAnchorFor: sessionId using: aMaReadStrategy

	| result |
	result _ MaReadResult byteArray:
		(repository anchorByteArrayUsing: 
			(self 
				ensureReadStrategy: aMaReadStrategy
				forSession: sessionId)).
	repository
		renewServerNotificationsIn: result
		for: sessionId.
	^result! !

!MagmaRepositoryController methodsFor: 'client requests' stamp: 'cmm 9/1/2003 19:21'!
refreshView

	self session abort.
	repository systemReadStrategy makeReadyForUseUsing: session serializer classIdManager! !

!MagmaRepositoryController methodsFor: 'private' stamp: ''!
repository

	^repository
! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 12/31/2002 12:54'!
repository: aMaObjectRepository

	repository _ aMaObjectRepository.
	aMaObjectRepository repositoryController: self! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 12/27/2002 15:55'!
requestCritical: aBlock
	"Coordinate access to all server elements."

	^requestInterruptGuard critical: aBlock! !

!MagmaRepositoryController methodsFor: 'private' stamp: ''!
serverSerializer
	"The serverSerializer is used for processing client requests.  It is also
	used during administration and possibly needed during initial creation of a
	repository."

	^serverSerializer
! !

!MagmaRepositoryController methodsFor: 'private' stamp: ''!
session
	"The server-side session is needed for largeCollection updates to be
	processed correctly.  When a client makes a change to a largeCollection, such
	as adding or removing an index, the systemRoot change is reflected upon
	the servers session refreshing its view of the database."

	^session
! !

!MagmaRepositoryController methodsFor: 'accessing' stamp: 'cmm 4/3/2003 22:59'!
sessionId

	^session id
! !

!MagmaRepositoryController methodsFor: 'client requests' stamp: 'cmm 10/22/2002 23:18'!
signalRequestGuard

	requestInterruptGuard signal! !

!MagmaRepositoryController methodsFor: 'client requests' stamp: ''!
sizeOfLargeCollection: oidInteger

	^(self largeCollectionManagerFor: oidInteger) size
! !

!MagmaRepositoryController methodsFor: 'client requests' stamp: 'cmm 12/26/2004 22:38'!
submitAll: aMaCommitPackage for: sessionId beginAnother: aBoolean

	^ repository
		submitAll: aMaCommitPackage
		for: sessionId
		beginAnother: aBoolean! !

!MagmaRepositoryController methodsFor: 'client requests' stamp: 'cmm 3/21/2005 15:54'!
upTo: anInteger oidsFrom: lowKey to: highKey startingAt: indexInteger inLargeCollection: collectionOid on: attributeSymbol using: aMaReadStrategy forSession: sessionId

	| segmentResult |

	segmentResult _ (self largeCollectionManagerFor: collectionOid)
		upTo: anInteger
		inSegmentFrom: lowKey
		to: highKey
		startingAt: indexInteger
		on: attributeSymbol.
	segmentResult
		objects:
			(segmentResult objects collect:
				[ :each | 
				repository
					byteArrayAt: each
					using:
						(self 
							ensureReadStrategy: aMaReadStrategy
							forSession: sessionId) ]).
	repository
		renewServerNotificationsIn: segmentResult
		for: sessionId.
	^ segmentResult! !

!MagmaRepositoryController methodsFor: 'client requests' stamp: 'cmm 4/30/2005 11:18'!
upTo: anInteger oidsFromIndex: startIndex inLargeCollection: collectionOid using: aMaReadStrategy forSession: sessionId

	| segmentResult readStrategy |
	segmentResult _ (self largeCollectionManagerFor: collectionOid)
		upTo: anInteger
		inSegmentFromIndex: (startIndex max: 1).
	readStrategy _
		self 
			ensureReadStrategy: aMaReadStrategy
			forSession: sessionId.
	segmentResult
		objects:
			(segmentResult objects collect:
				[ :each | 
				repository
					byteArrayAt: each
					using: readStrategy ]).
	repository
		renewServerNotificationsIn: segmentResult
		for: sessionId.
	^ segmentResult! !

!MagmaRepositoryController methodsFor: 'client/server required methods' stamp: 'cmm 9/19/2004 22:08'!
value: aMagmaRepositoryRequest

	^ self processRequest: aMagmaRepositoryRequest! !

!MagmaRepositoryController methodsFor: 'statistics' stamp: 'cmm 4/21/2003 16:02'!
version

	^repository version! !

MaObject subclass: #MagmaServerStatistics
	instanceVariableNames: 'startTime stopTime socketsQueuedStat requestsQueuedStat requestSizeStat requestProcessingTimeStat responseSizeStat stopwatch'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-API'!

!MagmaServerStatistics methodsFor: 'counting' stamp: 'cmm 11/16/2004 21:49'!
decrementRequestsQueued

	requestsQueuedStat last: requestsQueuedStat last - 1! !

!MagmaServerStatistics methodsFor: 'counting' stamp: 'cmm 11/16/2004 21:50'!
decrementSocketsQueued

	socketsQueuedStat last: socketsQueuedStat last - 1! !

!MagmaServerStatistics methodsFor: 'counting' stamp: 'cmm 11/16/2004 21:47'!
incrementRequestsQueued

	requestsQueuedStat addValue: 1 + requestsQueuedStat last! !

!MagmaServerStatistics methodsFor: 'counting' stamp: 'cmm 11/16/2004 21:47'!
incrementSocketsQueued

	socketsQueuedStat addValue: socketsQueuedStat last + 1! !

!MagmaServerStatistics methodsFor: 'initializing' stamp: 'cmm 11/16/2004 21:50'!
initialize
	super initialize.
	stopwatch _ MaStopwatch new.
	socketsQueuedStat _ MaStat new last: 0.
	requestsQueuedStat _ MaStat new last: 0.
	requestSizeStat _ MaStat new last: 0.
	requestProcessingTimeStat _ MaStat new last: 0.
	responseSizeStat _ MaStat new last: 0! !

!MagmaServerStatistics methodsFor: 'printing' stamp: 'cmm 2/5/2003 16:29'!
printOn: aStream

	super printOn: aStream.
	aStream
		cr; tab;
		nextPutAll: 'sockets queued: ';
		maPrintAbbreviated: socketsQueuedStat;
		cr; tab;
		nextPutAll: 'requests queued:  ';
		maPrintAbbreviated: requestsQueuedStat;
		cr; tab;
		nextPutAll: 'request sizes:  ';
		maPrintAbbreviated: requestSizeStat;
		cr; tab;
		nextPutAll: 'request processing time:  ';
		maPrintAbbreviated: requestProcessingTimeStat;
		cr; tab;
		nextPutAll: 'response sizes:  ';
		maPrintAbbreviated: responseSizeStat! !

!MagmaServerStatistics methodsFor: 'counting' stamp: 'cmm 11/16/2004 10:51'!
recordRequestSize: anInteger

	requestSizeStat addValue: anInteger! !

!MagmaServerStatistics methodsFor: 'counting' stamp: 'cmm 11/16/2004 10:51'!
recordResponseSize: anInteger

	responseSizeStat addValue: anInteger! !

!MagmaServerStatistics methodsFor: 'accessing' stamp: 'cmm 1/27/2004 16:20'!
requestProcessingTimeStat

	^requestProcessingTimeStat! !

!MagmaServerStatistics methodsFor: 'accessing' stamp: 'cmm 1/27/2004 16:20'!
requestSizeStat

	^requestSizeStat! !

!MagmaServerStatistics methodsFor: 'accessing' stamp: 'cmm 1/27/2004 16:20'!
requestsQueuedStat

	^requestsQueuedStat! !

!MagmaServerStatistics methodsFor: 'accessing' stamp: 'cmm 1/27/2004 16:20'!
responseSizeStat

	^responseSizeStat! !

!MagmaServerStatistics methodsFor: 'accessing' stamp: 'cmm 1/27/2004 16:20'!
socketsQueuedStat

	^socketsQueuedStat! !

!MagmaServerStatistics methodsFor: 'actions' stamp: 'cmm 1/5/2003 22:57'!
startRequestTimer

	stopwatch reset; start! !

!MagmaServerStatistics methodsFor: 'accessing' stamp: 'cmm 7/1/2004 00:34'!
startTime

	^ startTime! !

!MagmaServerStatistics methodsFor: 'accessing' stamp: 'cmm 10/6/2002 23:23'!
startTime: aMaCalendarTime

	startTime _ aMaCalendarTime! !

!MagmaServerStatistics methodsFor: 'actions' stamp: 'cmm 11/16/2004 10:52'!
stopRequestTimer

	stopwatch stop.
	requestProcessingTimeStat addValue: stopwatch time! !

!MagmaServerStatistics methodsFor: 'accessing' stamp: 'cmm 7/1/2004 00:34'!
stopTime

	^ stopTime! !

!MagmaServerStatistics methodsFor: 'accessing' stamp: 'cmm 10/7/2002 00:33'!
stopTime: aMaCalendarTime

	stopTime _ aMaCalendarTime! !

!MaSerializedGraphBuffer methodsFor: '*magma server' stamp: 'cmm 8/13/2002 21:27'!
decreaseBufferSizeBy: anInteger

	self bufferSize: self bufferSize - anInteger
! !

!MaByteObjectBuffer methodsFor: '*magma server' stamp: 'cmm 1/25/2005 14:07'!
appendGraphUsing: aMaReadStrategy into: aMaSerializedGraphBuffer currentDepth: currentDepth minDepth: minDepth with: alreadyAppendedSet filer: aMaObjectFiler

	"Nothing to do for a byte-object."
! !

!MaByteObjectBuffer methodsFor: '*magma server' stamp: 'cmm 5/1/2005 23:05'!
graphDo: oneArgBlock using: aMaReadStrategy for: aMagmaFileTraverser inDefinition: aBoolean

	"nothing to do for a Byte-object buffer."! !

!MaFixedObjectBuffer methodsFor: '*magma server' stamp: 'cmm 5/2/2005 12:05'!
graphDo: oneArgBlock using: aMaReadStrategy for: aMagmaFileTraverser inDefinition: aBoolean

	aMagmaFileTraverser
		graphFromPointers: self
		do: oneArgBlock
		using: aMaReadStrategy.
	(aBoolean not and: [ aMagmaFileTraverser isOidForLargeCollection: self oid ])
		ifTrue:
			[ aMagmaFileTraverser
				graphFromLargeCollection: self oid
				do: oneArgBlock
				using: aMaReadStrategy ]! !

!MagmaLargeCollectionChanges methodsFor: '*magma server' stamp: 'cmm 3/17/2005 22:08'!
createNewLargeCollectionUsing: aMaObjectRepository
	"Create the necessary support-files for this type of index."

	self subclassResponsibility! !

!MagmaLargeCollectionChanges methodsFor: '*magma server' stamp: 'cmm 3/16/2005 15:13'!
requiresServerRefresh

	^ self isNewCollection! !

!MagmaLargeCollectionChanges methodsFor: '*magma server' stamp: 'cmm 3/19/2005 11:07'!
suggestedFileExtension

	self subclassResponsibility! !

!MagmaSession class methodsFor: '*magma server-private' stamp: 'cmm 2/8/2005 21:44'!
newLocalLinkFor: aMagmaRepositoryController

	^ MaLocalServerLink localRequestServer: (MaLocalRequestServer processor: aMagmaRepositoryController)! !

!MagmaSession class methodsFor: '*magma server' stamp: 'cmm 10/26/2004 12:48'!
openLocal: pathString

	^ self repositoryController: (MagmaRepositoryController openedOn: pathString)! !

!MagmaSession class methodsFor: '*magma server' stamp: 'cmm 12/26/2004 23:21'!
repositoryController: aMagmaRepositoryController
	"Use this constructor only for *additional* local connections, i.e., two or more sessions connected to the same local repository.  Use the same MagmaRepositoryController you used to create the original."

	^ self link: (self newLocalLinkFor: aMagmaRepositoryController)! !

!MagmaSession methodsFor: '*magma server' stamp: 'cmm 2/15/2005 10:51'!
ensureOpen

	self isLocal ifFalse: [ MagmaUserError signal: 'Only available for local connections.' ].
	self isOpen ifFalse:
		[ self link: (self class newLocalLinkFor: (MagmaRepositoryController openedOn: self pathName)) ]! !

!MagmaSession methodsFor: '*magma server' stamp: 'cmm 1/13/2005 18:01'!
isOpen
	^ self isRemote 
	or: [ "local" self repositoryController isOpen ]! !

!MagmaSession methodsFor: '*magma server' stamp: 'cmm 1/13/2005 18:01'!
repositoryController

	^ link localRequestServer processor! !

!MaStorageObjectBuffer methodsFor: '*magma server' stamp: 'cmm 5/2/2005 13:49'!
graphDo: oneArgBlock using: aMaReadStrategy for: aMagmaFileTraverser inDefinition: aBoolean
	"Though I am an object stored in a different representation, that representation is fully part of the repository, so enumerate the buffers my MaSerializedGraphBuffers buffers."

	oneArgBlock value: self.
	self buffersDo:
		[ : eachBuffer |
		"The buffers of a storage-object are never assigned permanent oids.  But if already-persistent objects are referenced from the storage-object, they will be and should be traversed normally."
		(MaOidCalculator isOidForUserObject: eachBuffer oid)
			ifTrue:
				[ eachBuffer
					graphDo: oneArgBlock
					using: aMaReadStrategy
					for: aMagmaFileTraverser
					inDefinition: aBoolean ]
			ifFalse:
				[ oneArgBlock value: eachBuffer ] ]! !

PackageInfo subclass: #MagmaServerPackageInfo
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-Info'!

!MagmaServerPackageInfo methodsFor: 'as yet unclassified' stamp: 'cmm 10/31/2004 23:42'!
changesText

	^'
1.0gamma8
	- removed "allowNewClasses" feature.  This will be handled with real security later.
	- all administration functions are now either remote requests or no longer needed.
	- support for first-class MagmaRepositoryCodeBase.

1.0gamma7
	- Compatible with Squeak 3.7.
	- Better support for Symbols (thanks to Brent Pinkney!!).
	- fix for storing Exceptions in the repository (to accommodate changes in Squeak 3.6).
	- New Date index type for MagmaCollections.
	- New Float index type for MagmaCollections.
	- MagmaSessions cleaned from MethodContext receivers, allowing full garbage collection (thanks to Giovanni!!).
	- Critical bugs with MagmaCollections fixed and now pass the new, extra-stringent tests in the Tester package.
	
1.0gamma6
	- works with Squeak 3.6.  Doesn''t work with versions before 3.6.
	- fixed issue with servers that were running when the image was saved (functionality implemented in 1.0gamma5).  Fixed similar bug for saving with open local repository connections.
	- disconnect now cleans up the proxies for that session.  Please be patient, this can take a while if you have a lot.  Also now cleaning dynamic largeCollection attributes when disconnecting a session.
	- fixed bug with creating repository using an unqualified file name.
	- fixed an unnoticed reference to platform-specific code (ByteArray>>unsignedByteAt: uses FFI primitives).
	- improved the internal system-read-strategy.


1.0gamma5
	- Servers and clients can now survive image restarts.
	- added a new session preference (see MagmaPreferences), "signalProxyMaterialization" so you can capture MagmaNotifications and fine-tune your read strategies.
	- Another performance improvement!!  ReadStrategies are now cached on the server, instead of in your MagmaSession, improving read-efficiency when using read strategies because the strategy-object no longer needs to be serialized and transported for every proxy materialization.  Additionally, ReadStrategies for classes that don''t exist in the repository are no longer dropped.  Instead, they are remembered until you put those classes into the repository.
	- commit conflicts no longer provide the entire MagmaClientConnection object who conflicted with your session, just the userId.  This is due to the aforementioned cached readStrategy, no need to be sending that down on a commit conflict.
	- Since we have eradicated the "root trap" we can make replacing the root consistent with other Magma changes.  A transaction is now required.
	- several bug fixes.
'! !

!MagmaServerPackageInfo methodsFor: 'accessing' stamp: 'cmm 4/8/2004 21:30'!
prerequisitePackages
	^ #( 'Magma client' 'Ma time objects' )! !

!MagmaServerPackageInfo methodsFor: 'as yet unclassified' stamp: 'cmm 2/4/2003 18:12'!
readmeText

	^'This package is used to host a Magma repository.  For more information visit http://minnow.cc.gatech.edu/squeak/2665.'! !

!MagmaLargeCollection methodsFor: '*magma server' stamp: 'cmm 3/16/2005 12:53'!
newManagerUsing: aMagmaSession

	self subclassResponsibility! !

!MaCommitPackage methodsFor: '*magma server' stamp: 'cmm 4/5/2005 21:34'!
shouldServerRefresh
	"Anytime a largeCollection domain object changes, the server should
	refresh so we process all of the new large collections or their new
	indexes."

	^ allLargeCollectionChanges notNil
	and: [ allLargeCollectionChanges anySatisfy: [ :each | each requiresServerRefresh ] ]! !

!MagmaArray methodsFor: '*magma server' stamp: 'cmm 3/17/2005 21:33'!
newManagerUsing: aMaObjectRepository

	^ MagmaArrayManager
		collection: self
		repository: aMaObjectRepository! !

!MaObjectBuffer methodsFor: '*magma server' stamp: 'cmm 1/25/2005 14:58'!
appendGraphNode: oid into: aMaSerializedGraphBuffer using: aMaReadStrategy currentDepth: anInteger minDepth: minDepth with: alreadyAppendedSet filer: aMaObjectFiler

	| buffer |

	(alreadyAppendedSet includes: oid) ifTrue: [ ^self ].
	alreadyAppendedSet add: oid.
	buffer _ aMaObjectFiler
		appendObject: oid
		into: aMaSerializedGraphBuffer.
	buffer
		appendGraphUsing: aMaReadStrategy
		into: aMaSerializedGraphBuffer
		currentDepth: anInteger + 1
		minDepth: minDepth
		with: alreadyAppendedSet
		filer: aMaObjectFiler! !

!MaObjectBuffer methodsFor: '*magma server' stamp: 'cmm 1/25/2005 14:01'!
appendGraphUsing: aMaReadStrategy into: aMaSerializedGraphBuffer currentDepth: currentDepth minDepth: minDepth with: alreadyAppendedSet filer: aMaObjectFiler

	1
		to: self objectInstSize
		do:
			[ :varIndex | | oid |
			(MaOidCalculator isOidForOptimizedObject: (oid _ self maInstVarAt: varIndex))
				ifFalse:
					[ | newMinDepth |
					currentDepth < (
						newMinDepth _
							(aMaReadStrategy
								depthOfAttribute: varIndex
								onClassWithId: self classId
								version: self classVersion) + currentDepth
									max: minDepth)
										ifTrue:
											[ self
												appendGraphNode: oid
												into: aMaSerializedGraphBuffer
												using: aMaReadStrategy
												currentDepth: currentDepth
												minDepth: newMinDepth
												with: alreadyAppendedSet
												filer: aMaObjectFiler ] ] ]
! !

!MaObjectBuffer methodsFor: '*magma server' stamp: 'cmm 3/28/2005 21:53'!
asSentientBuffer
	"Answer the newwwwwww meeee!!  That's right!!  I've been injected
	with the ByteArray-of-truth and now know what kind of buffer
	I always wondered I was going to grow up to be."

	^self species
		createFromByteArray: byteArray
		startPos: startPos! !

!MaObjectBuffer methodsFor: '*magma server' stamp: 'cmm 4/27/2005 22:56'!
bufferPositionOfBytesToStore
	"Default 1-based position of the byte within each byteArray that we start storing to disk at."

	^ 5! !

!MaObjectBuffer methodsFor: '*magma server' stamp: 'cmm 5/1/2005 22:57'!
graphDo: oneArgBlock using: aMaReadStrategy for: aMagmaFileTraverser inDefinition: aBoolean

	self subclassResponsibility! !

!MaObjectBuffer methodsFor: '*magma server' stamp: 'cmm 11/24/2004 13:35'!
numberOfBytesToStore
	"Answer the number of bytes that should be stored for this buffer.  This is used by Magma, which doesn't need to store the oid since that is kept in the oid table."

	^self physicalSize - self bufferPositionOfBytesToStore + 1! !

!MagmaArrayChanges methodsFor: '*magma server' stamp: 'cmm 4/28/2005 20:40'!
createNewLargeCollectionUsing: aMaObjectRepository
	"Create the necessary support-files for this type of index."

	MaLargeArrayOfNumbers 
		create: 
			(aMaObjectRepository 
				collectionFilenameFor: collectionOid
				extension: self suggestedFileExtension)
		bitSize: MaObjectFiler maxDbSize! !

!MagmaArrayChanges methodsFor: '*magma server' stamp: 'cmm 4/28/2005 20:40'!
suggestedFileExtension

	^ MaLargeArrayOfNumbers suggestedFileExtension ! !

!MaLargeArrayOfNumbers methodsFor: '*magma server' stamp: 'cmm 3/22/2005 18:07'!
do: aBlock for: aMagmaFileTraverser using: aMaReadStrategy

	1 to: self size do:
		[ : x |  | oid |
		(oid _ self at: x) ~= 0
			ifTrue:
				[ aMagmaFileTraverser
					graphFrom: oid
					do: aBlock
					using: aMaReadStrategy ] ]! !

!MaReadStrategy methodsFor: '*magma server' stamp: 'cmm 1/26/2005 12:22'!
depthOfAttribute: attributeIndex onClassWithId: classId version: versionNumber
	"Answer the depth that the attribute at attributeIndex should be traversed into *from the current depth*."

	^depths
		maAt: { classId. versionNumber }
		ifPresent:
			[ :depth |
			depth isInteger
				ifTrue: [ depth ]
				ifFalse: [ depth at: attributeIndex ifAbsent: [ 0 ] ] ]
		ifAbsent: [ 0 ]! !

!MaVariableBuffer methodsFor: '*magma server' stamp: 'cmm 5/2/2005 12:04'!
graphDo: oneArgBlock using: aMaReadStrategy for: aMagmaFileTraverser inDefinition: aBoolean

	aMagmaFileTraverser
		graphFromPointers: self
		do: oneArgBlock
		using: aMaReadStrategy! !

MaRequestServerConsole subclass: #MagmaServerConsole
	instanceVariableNames: 'server controller statistics'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-API'!

!MagmaServerConsole methodsFor: 'accessing' stamp: 'cmm 12/18/2002 21:52'!
controller

	^controller! !

!MagmaServerConsole methodsFor: 'initializing' stamp: 'cmm 12/24/2002 15:50'!
initialize
	super initialize.
	statistics := MagmaServerStatistics new! !

!MagmaServerConsole methodsFor: 'notifications' stamp: 'cmm 1/4/2003 17:17'!
note: stringMessage

	Transcript cr; show: stringMessage! !

!MagmaServerConsole methodsFor: 'notifications' stamp: 'cmm 1/5/2003 22:56'!
noteProcessingRequest: yourRequestObject

	statistics startRequestTimer! !

!MagmaServerConsole methodsFor: 'notifications' stamp: 'cmm 1/5/2003 22:53'!
noteRequestDequeued: requestByteArray for: aSocket
	"Indicates aSocket has delivered a ByteArray request that is about to be processed by my server."

	statistics decrementRequestsQueued! !

!MagmaServerConsole methodsFor: 'notifications' stamp: 'cmm 2/4/2003 12:36'!
noteRequestProcessed: requestObject

	statistics stopRequestTimer! !

!MagmaServerConsole methodsFor: 'notifications' stamp: 'cmm 1/5/2003 22:44'!
noteRequestQueued: aByteArray for: aSocket
	"Indicates aSocket has delivered aByteArray as a request that it wants processed and the response returned."

	statistics
		incrementRequestsQueued;
		recordRequestSize: aByteArray size! !

!MagmaServerConsole methodsFor: 'notifications' stamp: 'cmm 2/4/2003 13:09'!
noteResponseSent: aByteArray to: aSocket
	"Successfully sent the aByteArray to aSocket."

	statistics recordResponseSize: aByteArray size! !

!MagmaServerConsole methodsFor: 'notifications' stamp: 'cmm 4/30/2003 17:44'!
noteServerShutdown

	controller close.
	statistics stopTime: MaCalendarTime now.
	self note: controller printString, ' closed.'! !

!MagmaServerConsole methodsFor: 'notifications' stamp: 'cmm 1/5/2003 18:19'!
noteSocketDequeued: aSocket

	statistics decrementSocketsQueued! !

!MagmaServerConsole methodsFor: 'notifications' stamp: 'cmm 10/7/2002 22:06'!
noteSocketQueued: aSocket

	statistics incrementSocketsQueued! !

!MagmaServerConsole methodsFor: 'start / stop' stamp: 'cmm 2/20/2005 19:39'!
open: pathName processOn: portInteger

	controller _ MagmaRepositoryController open: pathName.
	server _ (MaTcpRequestServer protocol: controller protocol) console: self.
	statistics startTime: MaCalendarTime now.
	server 
		processOn: portInteger
		using: [ :eachRequest | controller processRequest: eachRequest ].
	^self! !

!MagmaServerConsole methodsFor: 'start / stop' stamp: 'cmm 12/17/2002 14:44'!
openReadOnly: fileNameString processOn: portInteger

	self
		open: fileNameString
		processOn: portInteger.
	controller beReadOnly! !

!MagmaServerConsole methodsFor: 'accessing' stamp: 'cmm 3/3/2005 10:10'!
server

	^ server! !

!MagmaServerConsole methodsFor: 'start / stop' stamp: 'cmm 10/7/2002 00:42'!
shutdown

	server shutdown! !

!MagmaServerConsole methodsFor: 'accessing' stamp: 'cmm 1/27/2004 15:55'!
statistics

	^statistics! !

!MaBasicReadStrategy methodsFor: '*magma server' stamp: 'cmm 1/26/2005 17:28'!
depthOfAttribute: attributeIndex onClassWithId: classId version: versionNumber
	"Answer the delta-depth from the current depth to read.  This is distinct from my minimumDepth, which was used at the start of this graph-read."

	^ 0! !

!MaOidCalculator class methodsFor: '*magma server' stamp: 'cmm 12/2/2004 13:44'!
indexPositionForOid: anInteger

	^ anInteger - self firstUserObjectOid! !

!MaOidCalculator class methodsFor: '*magma server' stamp: 'cmm 8/13/2002 23:57'!
isOidForOptimizedObject: oidInteger
	"Answer whether oidInteger is an oid for anObject which will never require its own buffer.  These include optimized Integers, nil, true, false, and all Characters."

	^(self isOidForNewObject: oidInteger) not
		and: [ (self isOidForUserObject: oidInteger) not ]! !

!MaOidCalculator class methodsFor: '*magma server' stamp: 'cmm 11/13/2002 23:05'!
isOidForUserObject: anInteger
	"A 'user' object is a persistent object stored in a Magma database that does not have a special oid, such as SmallIntegers, Floats, Booleans, etc. have."

	^anInteger
		between: self firstUserObjectOid
		and: self lastUserObjectOid! !

!MaOidCalculator class methodsFor: '*magma server' stamp: 'cmm 4/21/2005 15:41'!
lastUserObjectOid

	^ 281468533259712   "self oidForSmallestFloat - 1"! !

!MaOidCalculator class methodsFor: '*magma server' stamp: 'cmm 12/2/2004 13:32'!
oidForFilePositionIndexPosition: anInteger

	^ self firstUserObjectOid + anInteger! !
MagmaRepositoryController initialize!
